from __future__ import annotations

import json
from dataclasses import dataclass, field, replace
from typing import Any, Callable, Literal

from meshagent.api.messaging import JsonContent, TextContent

from .messages import (
    AGENT_EVENT_FILE_CONTENT_DELTA,
    AGENT_EVENT_FILE_CONTENT_ENDED,
    AGENT_EVENT_FILE_CONTENT_STARTED,
    AGENT_EVENT_REASONING_CONTENT_DELTA,
    AGENT_EVENT_REASONING_CONTENT_ENDED,
    AGENT_EVENT_REASONING_CONTENT_STARTED,
    AGENT_EVENT_TEXT_CONTENT_DELTA,
    AGENT_EVENT_TEXT_CONTENT_ENDED,
    AGENT_EVENT_TEXT_CONTENT_STARTED,
    AGENT_EVENT_TOOL_CALL_PENDING,
    AGENT_EVENT_TOOL_CALL_IN_PROGRESS,
    AGENT_EVENT_TOOL_CALL_LOG_DELTA,
    AGENT_EVENT_TOOL_CALL_ENDED,
    AGENT_EVENT_TOOL_CALL_STARTED,
    AgentError,
    AgentFileContentDelta,
    AgentFileContentEnded,
    AgentFileContentStarted,
    AgentMessage,
    AgentReasoningContentDelta,
    AgentReasoningContentEnded,
    AgentReasoningContentStarted,
    AgentTextContentDelta,
    AgentTextContentEnded,
    AgentTextContentStarted,
    AgentToolCallPending,
    AgentToolCallInProgress,
    AgentToolCallLogDelta,
    AgentToolCallLogLine,
    AgentToolCallEnded,
    AgentToolCallStarted,
)

AgentEventCallback = Callable[[AgentMessage], None]
FunctionToolNameResolver = Callable[[str], tuple[str, str] | None]
_ContentKind = Literal["file", "reasoning", "text"]
_DEFERRED_HANDLER_RESULT_TOOL_TYPES = {
    "apply_patch_call",
    "computer_call",
    "local_shell_call",
    "shell_call",
}
_HANDLER_RESULT_TOOL_TYPES = _DEFERRED_HANDLER_RESULT_TOOL_TYPES | {"function_call"}


def _as_dict(value: Any) -> dict[str, Any] | None:
    if isinstance(value, dict):
        return value
    return None


def _as_int(value: Any) -> int | None:
    if isinstance(value, bool):
        return None
    if isinstance(value, int):
        return value
    try:
        return int(value)
    except Exception:
        return None


def _as_str(value: Any) -> str | None:
    if isinstance(value, str):
        stripped = value.strip()
        return stripped if stripped != "" else None
    return None


def _as_text(value: Any) -> str | None:
    if isinstance(value, str):
        return value
    return None


def _parse_tool_arguments(value: Any) -> dict[str, Any] | None:
    if value is None:
        return None
    if isinstance(value, dict):
        return value
    if isinstance(value, str):
        stripped = value.strip()
        if stripped == "":
            return None
        try:
            parsed = json.loads(stripped)
        except json.JSONDecodeError:
            return {"raw": stripped}
        if isinstance(parsed, dict):
            return parsed
        return {"value": parsed}
    return None


def _parse_tool_log_lines(value: Any) -> list[AgentToolCallLogLine]:
    if not isinstance(value, list):
        return []

    parsed_lines: list[AgentToolCallLogLine] = []
    for entry in value:
        if not isinstance(entry, dict):
            continue

        source = _as_str(entry.get("source"))
        text = _as_text(entry.get("text"))
        if source not in {"stdout", "stderr"} or text is None:
            continue

        parsed_lines.append(AgentToolCallLogLine(source=source, text=text))

    return parsed_lines


def _data_url(*, mime_type: str, data: str) -> str:
    return f"data:{mime_type};base64,{data}"


def _file_url_from_payload(
    *,
    payload: dict[str, Any],
    provider: Literal["anthropic", "openai"],
) -> str | None:
    direct_url_keys = ("url", "file_url")
    for key in direct_url_keys:
        url = _as_str(payload.get(key))
        if url is not None:
            return url

    nested_url_keys = ("image_url", "file", "source")
    for key in nested_url_keys:
        nested = _as_dict(payload.get(key))
        if nested is None:
            continue
        url = _as_str(nested.get("url"))
        if url is not None:
            return url
        nested_type = _as_str(nested.get("type"))
        if nested_type == "base64":
            media_type = _as_str(nested.get("media_type")) or _as_str(
                nested.get("mime_type")
            )
            data = _as_str(nested.get("data"))
            if media_type is not None and data is not None:
                return _data_url(mime_type=media_type, data=data)

    direct_base64_keys = (
        "partial_image_b64",
        "image_base64",
        "image_b64",
        "b64_json",
        "data",
    )
    media_type = _as_str(payload.get("mime_type")) or _as_str(payload.get("media_type"))
    for key in direct_base64_keys:
        data = _as_str(payload.get(key))
        if data is not None and media_type is not None:
            return _data_url(mime_type=media_type, data=data)

    file_id = _as_str(payload.get("file_id"))
    if file_id is not None:
        return f"{provider}://file/{file_id}"

    return None


def _part_text(part: dict[str, Any]) -> str:
    text = _as_text(part.get("text"))
    if text is not None:
        return text

    refusal = _as_text(part.get("refusal"))
    if refusal is not None:
        return refusal

    return ""


def _content_kind_from_part(part: dict[str, Any]) -> _ContentKind | None:
    part_type = (_as_str(part.get("type")) or "").lower()

    if part_type in {"output_text", "refusal", "text"}:
        return "text"

    if part_type in {"reasoning_text", "summary_text"} or part_type.startswith(
        "reasoning"
    ):
        return "reasoning"

    if part_type in {
        "document",
        "file",
        "image",
        "input_image",
        "output_file",
        "output_image",
    }:
        return "file"

    if _file_url_from_payload(payload=part, provider="openai") is not None:
        return "file"

    return None


def _content_from_tool_item(item: dict[str, Any]):
    result = item.get("result")
    if isinstance(result, dict):
        return JsonContent(json=result)
    if isinstance(result, list):
        return JsonContent(json={"result": result})
    if isinstance(result, str) and result.strip() != "":
        return TextContent(text=result)

    output = item.get("output")
    if isinstance(output, dict):
        return JsonContent(json=output)
    if isinstance(output, list):
        return JsonContent(json={"output": output})
    if isinstance(output, str) and output.strip() != "":
        return TextContent(text=output)

    results = item.get("results")
    if isinstance(results, dict):
        return JsonContent(json=results)
    if isinstance(results, list):
        return JsonContent(json={"results": results})

    return None


def _content_from_handler_result(value: Any):
    if isinstance(value, dict):
        return JsonContent(json=value)
    if isinstance(value, list):
        return JsonContent(json={"result": value})
    if isinstance(value, str) and value.strip() != "":
        return TextContent(text=value)
    return None


@dataclass(frozen=True, slots=True)
class _ToolCallInfo:
    item_id: str
    toolkit: str
    tool: str
    item_type: str | None = None
    call_id: str | None = None
    arguments: dict[str, Any] | None = None
    error: AgentError | None = None
    result: TextContent | JsonContent | None = None


def _filtered_tool_arguments(
    *,
    item: dict[str, Any],
    excluded_keys: set[str],
) -> dict[str, Any] | None:
    arguments: dict[str, Any] = {}
    for key, value in item.items():
        if key in excluded_keys or value is None:
            continue
        arguments[key] = value
    return arguments or None


def _openai_tool_call_info(
    item: dict[str, Any],
    *,
    function_tool_name_resolver: FunctionToolNameResolver | None = None,
) -> _ToolCallInfo | None:
    item_type = _as_str(item.get("type"))
    item_id = _as_str(item.get("id"))
    if item_type is None or item_id is None:
        return None

    status = _as_str(item.get("status"))
    error_text = _as_str(item.get("error"))
    error = None
    if error_text is not None:
        error = AgentError(message=error_text, code=status or "tool_call_failed")

    if item_type == "function_call":
        safe_name = _as_str(item.get("name"))
        toolkit_name = "function"
        tool_name = safe_name or "function"
        if function_tool_name_resolver is not None and safe_name is not None:
            resolved = function_tool_name_resolver(safe_name)
            if resolved is not None:
                toolkit_name, tool_name = resolved
        return _ToolCallInfo(
            item_id=item_id,
            toolkit=toolkit_name,
            tool=tool_name,
            item_type=item_type,
            call_id=_as_str(item.get("call_id")),
            arguments=_parse_tool_arguments(item.get("arguments")),
            error=error,
            result=_content_from_tool_item(item),
        )

    if item_type == "mcp_call":
        return _ToolCallInfo(
            item_id=item_id,
            toolkit=_as_str(item.get("server_label")) or "mcp",
            tool=_as_str(item.get("name")) or "call",
            item_type=item_type,
            call_id=_as_str(item.get("call_id")),
            arguments=_parse_tool_arguments(item.get("arguments")),
            error=error,
            result=_content_from_tool_item(item),
        )

    if item_type == "mcp_list_tools":
        return _ToolCallInfo(
            item_id=item_id,
            toolkit=_as_str(item.get("server_label")) or "mcp",
            tool="list_tools",
            item_type=item_type,
            call_id=_as_str(item.get("call_id")),
            arguments=_filtered_tool_arguments(
                item=item,
                excluded_keys={"error", "id", "output", "status", "type"},
            ),
            error=error,
            result=_content_from_tool_item(item),
        )

    if not item_type.endswith("_call"):
        return None

    return _ToolCallInfo(
        item_id=item_id,
        toolkit="openai",
        tool=item_type.removesuffix("_call"),
        item_type=item_type,
        call_id=_as_str(item.get("call_id")),
        arguments=_filtered_tool_arguments(
            item=item,
            excluded_keys={
                "call_id",
                "error",
                "id",
                "output",
                "result",
                "results",
                "status",
                "type",
            },
        ),
        error=error,
        result=_content_from_tool_item(item),
    )


def _anthropic_tool_call_info(
    *,
    block: dict[str, Any],
    item_id: str,
    function_tool_name_resolver: FunctionToolNameResolver | None = None,
) -> _ToolCallInfo | None:
    block_type = _as_str(block.get("type"))
    if block_type is None:
        return None

    if block_type == "tool_use":
        safe_name = _as_str(block.get("name"))
        toolkit_name = "function"
        tool_name = safe_name or "tool"
        if function_tool_name_resolver is not None and safe_name is not None:
            resolved = function_tool_name_resolver(safe_name)
            if resolved is not None:
                toolkit_name, tool_name = resolved
        return _ToolCallInfo(
            item_id=item_id,
            toolkit=toolkit_name,
            tool=tool_name,
            arguments=_parse_tool_arguments(block.get("input")),
        )

    if block_type == "mcp_tool_use":
        return _ToolCallInfo(
            item_id=item_id,
            toolkit=(
                _as_str(block.get("server_name"))
                or _as_str(block.get("server_label"))
                or "mcp"
            ),
            tool=_as_str(block.get("name")) or "tool",
            arguments=_parse_tool_arguments(
                block.get("input")
                if block.get("input") is not None
                else block.get("arguments")
            ),
        )

    if block_type.endswith("_tool_use"):
        return _ToolCallInfo(
            item_id=item_id,
            toolkit="anthropic",
            tool=block_type.removesuffix("_tool_use"),
            arguments=_filtered_tool_arguments(
                item=block,
                excluded_keys={"id", "name", "type"},
            ),
        )

    return None


@dataclass(slots=True)
class _AgentMessageEmitter:
    turn_id: str
    thread_id: str
    callback: AgentEventCallback
    _started_content: set[tuple[_ContentKind, str]] = field(default_factory=set)
    _ended_content: set[tuple[_ContentKind, str]] = field(default_factory=set)
    _content_with_data: set[tuple[_ContentKind, str]] = field(default_factory=set)
    _pending_tool_calls: dict[str, _ToolCallInfo] = field(default_factory=dict)
    _in_progress_tool_calls: dict[str, _ToolCallInfo] = field(default_factory=dict)
    _started_tool_calls: dict[str, _ToolCallInfo] = field(default_factory=dict)
    _ended_tool_calls: dict[str, _ToolCallInfo] = field(default_factory=dict)

    def _content_key(
        self, *, kind: _ContentKind, item_id: str
    ) -> tuple[_ContentKind, str]:
        return kind, item_id

    def _ensure_content_started(self, *, kind: _ContentKind, item_id: str) -> None:
        key = self._content_key(kind=kind, item_id=item_id)
        if key in self._started_content:
            return

        self._started_content.add(key)
        if kind == "text":
            self.callback(
                AgentTextContentStarted(
                    type=AGENT_EVENT_TEXT_CONTENT_STARTED,
                    thread_id=self.thread_id,
                    turn_id=self.turn_id,
                    item_id=item_id,
                )
            )
            return

        if kind == "reasoning":
            self.callback(
                AgentReasoningContentStarted(
                    type=AGENT_EVENT_REASONING_CONTENT_STARTED,
                    thread_id=self.thread_id,
                    turn_id=self.turn_id,
                    item_id=item_id,
                )
            )
            return

        self.callback(
            AgentFileContentStarted(
                type=AGENT_EVENT_FILE_CONTENT_STARTED,
                thread_id=self.thread_id,
                turn_id=self.turn_id,
                item_id=item_id,
            )
        )

    def _mark_content_has_data(self, *, kind: _ContentKind, item_id: str) -> None:
        self._content_with_data.add(self._content_key(kind=kind, item_id=item_id))

    def has_content_data(self, *, kind: _ContentKind, item_id: str) -> bool:
        return self._content_key(kind=kind, item_id=item_id) in self._content_with_data

    def emit_text_delta(self, *, item_id: str, text: str) -> None:
        if text == "":
            return
        self._ensure_content_started(kind="text", item_id=item_id)
        self._mark_content_has_data(kind="text", item_id=item_id)
        self.callback(
            AgentTextContentDelta(
                type=AGENT_EVENT_TEXT_CONTENT_DELTA,
                thread_id=self.thread_id,
                turn_id=self.turn_id,
                item_id=item_id,
                text=text,
            )
        )

    def emit_reasoning_delta(self, *, item_id: str, text: str) -> None:
        if text == "":
            return
        self._ensure_content_started(kind="reasoning", item_id=item_id)
        self._mark_content_has_data(kind="reasoning", item_id=item_id)
        self.callback(
            AgentReasoningContentDelta(
                type=AGENT_EVENT_REASONING_CONTENT_DELTA,
                thread_id=self.thread_id,
                turn_id=self.turn_id,
                item_id=item_id,
                text=text,
            )
        )

    def emit_file_delta(self, *, item_id: str, url: str | None) -> None:
        if url is None:
            return
        self._ensure_content_started(kind="file", item_id=item_id)
        self._mark_content_has_data(kind="file", item_id=item_id)
        self.callback(
            AgentFileContentDelta(
                type=AGENT_EVENT_FILE_CONTENT_DELTA,
                thread_id=self.thread_id,
                turn_id=self.turn_id,
                item_id=item_id,
                url=url,
            )
        )

    def emit_text_started(self, *, item_id: str) -> None:
        self._ensure_content_started(kind="text", item_id=item_id)

    def emit_reasoning_started(self, *, item_id: str) -> None:
        self._ensure_content_started(kind="reasoning", item_id=item_id)

    def emit_file_started(self, *, item_id: str) -> None:
        self._ensure_content_started(kind="file", item_id=item_id)

    def emit_text_ended(self, *, item_id: str) -> None:
        key = self._content_key(kind="text", item_id=item_id)
        if key in self._ended_content:
            return
        self._ensure_content_started(kind="text", item_id=item_id)
        self._ended_content.add(key)
        self.callback(
            AgentTextContentEnded(
                type=AGENT_EVENT_TEXT_CONTENT_ENDED,
                thread_id=self.thread_id,
                turn_id=self.turn_id,
                item_id=item_id,
            )
        )

    def emit_reasoning_ended(self, *, item_id: str) -> None:
        key = self._content_key(kind="reasoning", item_id=item_id)
        if key in self._ended_content:
            return
        self._ensure_content_started(kind="reasoning", item_id=item_id)
        self._ended_content.add(key)
        self.callback(
            AgentReasoningContentEnded(
                type=AGENT_EVENT_REASONING_CONTENT_ENDED,
                thread_id=self.thread_id,
                turn_id=self.turn_id,
                item_id=item_id,
            )
        )

    def emit_file_ended(self, *, item_id: str) -> None:
        key = self._content_key(kind="file", item_id=item_id)
        if key in self._ended_content:
            return
        self._ensure_content_started(kind="file", item_id=item_id)
        self._ended_content.add(key)
        self.callback(
            AgentFileContentEnded(
                type=AGENT_EVENT_FILE_CONTENT_ENDED,
                thread_id=self.thread_id,
                turn_id=self.turn_id,
                item_id=item_id,
            )
        )

    def emit_tool_pending(self, *, info: _ToolCallInfo) -> None:
        existing = self._pending_tool_calls.get(info.item_id)
        if (
            existing is not None
            and existing.toolkit == info.toolkit
            and existing.tool == info.tool
            and existing.arguments == info.arguments
        ):
            return
        self._pending_tool_calls[info.item_id] = info
        self.callback(
            AgentToolCallPending(
                type=AGENT_EVENT_TOOL_CALL_PENDING,
                thread_id=self.thread_id,
                turn_id=self.turn_id,
                item_id=info.item_id,
                toolkit=info.toolkit,
                tool=info.tool,
                arguments=info.arguments,
            )
        )

    def emit_tool_in_progress(self, *, info: _ToolCallInfo) -> None:
        existing = self._in_progress_tool_calls.get(info.item_id)
        if (
            existing is not None
            and existing.toolkit == info.toolkit
            and existing.tool == info.tool
            and existing.arguments == info.arguments
        ):
            return
        self._pending_tool_calls.pop(info.item_id, None)
        self._in_progress_tool_calls[info.item_id] = info
        self.callback(
            AgentToolCallInProgress(
                type=AGENT_EVENT_TOOL_CALL_IN_PROGRESS,
                thread_id=self.thread_id,
                turn_id=self.turn_id,
                item_id=info.item_id,
                toolkit=info.toolkit,
                tool=info.tool,
                arguments=info.arguments,
            )
        )

    def emit_tool_started(self, *, info: _ToolCallInfo) -> None:
        existing = self._started_tool_calls.get(info.item_id)
        if (
            existing is not None
            and existing.toolkit == info.toolkit
            and existing.tool == info.tool
            and existing.arguments == info.arguments
        ):
            return
        self._pending_tool_calls.pop(info.item_id, None)
        self._in_progress_tool_calls.pop(info.item_id, None)
        self._started_tool_calls[info.item_id] = info
        self.callback(
            AgentToolCallStarted(
                type=AGENT_EVENT_TOOL_CALL_STARTED,
                thread_id=self.thread_id,
                turn_id=self.turn_id,
                item_id=info.item_id,
                toolkit=info.toolkit,
                tool=info.tool,
                arguments=info.arguments,
            )
        )

    def emit_tool_log_delta(
        self, *, item_id: str, lines: list[AgentToolCallLogLine]
    ) -> None:
        if len(lines) == 0:
            return
        self.callback(
            AgentToolCallLogDelta(
                type=AGENT_EVENT_TOOL_CALL_LOG_DELTA,
                thread_id=self.thread_id,
                turn_id=self.turn_id,
                item_id=item_id,
                lines=lines,
            )
        )

    def emit_tool_ended(self, *, info: _ToolCallInfo) -> None:
        self.emit_tool_started(info=info)
        existing = self._ended_tool_calls.get(info.item_id)
        if (
            existing is not None
            and existing.result == info.result
            and existing.error == info.error
        ):
            return
        self._pending_tool_calls.pop(info.item_id, None)
        self._in_progress_tool_calls.pop(info.item_id, None)
        self._started_tool_calls.pop(info.item_id, None)
        self._ended_tool_calls[info.item_id] = info
        self.callback(
            AgentToolCallEnded(
                type=AGENT_EVENT_TOOL_CALL_ENDED,
                thread_id=self.thread_id,
                turn_id=self.turn_id,
                item_id=info.item_id,
                result=info.result,
                error=info.error,
            )
        )


@dataclass(slots=True)
class _OpenAIAgentEventPublisher:
    emitter: _AgentMessageEmitter
    function_tool_name_resolver: FunctionToolNameResolver | None = None
    custom_event_callback: Callable[[dict[str, Any]], None] | None = None
    _active_response_id: str | None = None
    _output_item_ids: dict[int, str] = field(default_factory=dict)
    _pending_handler_tool_calls: dict[str, _ToolCallInfo] = field(default_factory=dict)
    _finished_handler_tool_call_ids: set[str] = field(default_factory=set)

    def set_function_tool_name_resolver(
        self,
        resolver: FunctionToolNameResolver | None,
    ) -> None:
        self.function_tool_name_resolver = resolver

    def _track_response_boundary(self, *, event: dict[str, Any]) -> None:
        response = _as_dict(event.get("response"))
        if response is None:
            return

        response_id = _as_str(response.get("id"))
        if response_id is None:
            return

        if self._active_response_id is None:
            self._active_response_id = response_id
            return

        if self._active_response_id == response_id:
            return

        self._active_response_id = response_id
        self._output_item_ids.clear()

    def _item_id_from_event(self, *, event: dict[str, Any]) -> str | None:
        output_index = _as_int(event.get("output_index"))
        item_id = _as_str(event.get("item_id"))

        item = _as_dict(event.get("item"))
        if item is not None and item_id is None:
            item_id = _as_str(item.get("id"))

        if output_index is not None:
            return self._mapped_output_item_id(
                output_index=output_index,
                item_id=item_id,
            )

        return item_id

    def _mapped_output_item_id(
        self,
        *,
        output_index: int,
        item_id: str | None = None,
    ) -> str:
        existing = self._output_item_ids.get(output_index)
        if existing is not None:
            return existing

        if item_id is not None:
            self._output_item_ids[output_index] = item_id
            return item_id

        synthesized = f"output:{output_index}"
        self._output_item_ids[output_index] = synthesized
        return synthesized

    def _record_output_item(
        self, *, event: dict[str, Any], item: dict[str, Any]
    ) -> None:
        output_index = _as_int(event.get("output_index"))
        item_id = _as_str(item.get("id"))
        if output_index is None or item_id is None:
            return
        self._mapped_output_item_id(output_index=output_index, item_id=item_id)

    def _finish_part_from_snapshot(self, *, item_id: str, part: dict[str, Any]) -> None:
        content_kind = _content_kind_from_part(part)
        if content_kind == "text":
            text = _part_text(part)
            if text != "" and not self.emitter.has_content_data(
                kind="text", item_id=item_id
            ):
                self.emitter.emit_text_delta(item_id=item_id, text=text)
            self.emitter.emit_text_ended(item_id=item_id)
            return

        if content_kind == "reasoning":
            text = _part_text(part)
            if text != "" and not self.emitter.has_content_data(
                kind="reasoning", item_id=item_id
            ):
                self.emitter.emit_reasoning_delta(item_id=item_id, text=text)
            self.emitter.emit_reasoning_ended(item_id=item_id)
            return

        if content_kind == "file":
            url = _file_url_from_payload(payload=part, provider="openai")
            if not self.emitter.has_content_data(kind="file", item_id=item_id):
                self.emitter.emit_file_delta(item_id=item_id, url=url)
            self.emitter.emit_file_ended(item_id=item_id)

    def _on_output_item(self, *, event: dict[str, Any], completed: bool) -> None:
        item = _as_dict(event.get("item"))
        if item is None:
            return

        self._record_output_item(event=event, item=item)

        tool_info = _openai_tool_call_info(
            item,
            function_tool_name_resolver=self.function_tool_name_resolver,
        )
        if tool_info is not None:
            if tool_info.item_id in self._finished_handler_tool_call_ids:
                return

            if tool_info.item_type in _HANDLER_RESULT_TOOL_TYPES:
                self._pending_handler_tool_calls[tool_info.item_id] = tool_info
                self.emitter.emit_tool_pending(info=tool_info)
                return

            if completed:
                self.emitter.emit_tool_ended(info=tool_info)
            else:
                self.emitter.emit_tool_started(info=tool_info)
            return

        item_type = _as_str(item.get("type"))
        item_id = self._item_id_from_event(event=event)
        if item_type == "reasoning" and item_id is not None:
            if not completed:
                self.emitter.emit_reasoning_started(item_id=item_id)
                return

            if not self.emitter.has_content_data(kind="reasoning", item_id=item_id):
                for bucket_name in ("summary", "content"):
                    bucket = item.get(bucket_name)
                    if not isinstance(bucket, list):
                        continue
                    for part in bucket:
                        if not isinstance(part, dict):
                            continue
                        text = _part_text(part)
                        if text != "":
                            self.emitter.emit_reasoning_delta(
                                item_id=item_id, text=text
                            )
            self.emitter.emit_reasoning_ended(item_id=item_id)
            return

        if item_type == "message" and completed and item_id is not None:
            content = item.get("content")
            if isinstance(content, list):
                for part in content:
                    if isinstance(part, dict):
                        self._finish_part_from_snapshot(item_id=item_id, part=part)

    def _on_content_part_added(self, *, event: dict[str, Any]) -> None:
        part = _as_dict(event.get("part"))
        item_id = self._item_id_from_event(event=event)
        if part is None or item_id is None:
            return

        content_kind = _content_kind_from_part(part)
        if content_kind == "text":
            self.emitter.emit_text_started(item_id=item_id)
            self.emitter.emit_text_delta(item_id=item_id, text=_part_text(part))
            return

        if content_kind == "reasoning":
            self.emitter.emit_reasoning_started(item_id=item_id)
            self.emitter.emit_reasoning_delta(item_id=item_id, text=_part_text(part))
            return

        if content_kind == "file":
            self.emitter.emit_file_started(item_id=item_id)
            self.emitter.emit_file_delta(
                item_id=item_id,
                url=_file_url_from_payload(payload=part, provider="openai"),
            )

    def _on_content_part_done(self, *, event: dict[str, Any]) -> None:
        part = _as_dict(event.get("part"))
        item_id = self._item_id_from_event(event=event)
        if part is None or item_id is None:
            return

        self._finish_part_from_snapshot(item_id=item_id, part=part)

    def _on_text_delta(self, *, event: dict[str, Any]) -> None:
        item_id = self._item_id_from_event(event=event)
        if item_id is None:
            return
        self.emitter.emit_text_delta(
            item_id=item_id, text=_as_text(event.get("delta")) or ""
        )

    def _on_text_done(self, *, event: dict[str, Any], field_name: str) -> None:
        item_id = self._item_id_from_event(event=event)
        if item_id is None:
            return

        final_text = _as_text(event.get(field_name)) or ""
        if final_text != "" and not self.emitter.has_content_data(
            kind="text", item_id=item_id
        ):
            self.emitter.emit_text_delta(item_id=item_id, text=final_text)
        self.emitter.emit_text_ended(item_id=item_id)

    def _on_reasoning_delta(self, *, event: dict[str, Any], field_name: str) -> None:
        item_id = self._item_id_from_event(event=event)
        if item_id is None:
            return
        self.emitter.emit_reasoning_delta(
            item_id=item_id,
            text=_as_text(event.get(field_name)) or "",
        )

    def _on_reasoning_done(
        self, *, event: dict[str, Any], field_name: str | None = None
    ) -> None:
        item_id = self._item_id_from_event(event=event)
        if item_id is None:
            return

        if field_name is not None:
            final_text = _as_text(event.get(field_name)) or ""
            if final_text != "" and not self.emitter.has_content_data(
                kind="reasoning", item_id=item_id
            ):
                self.emitter.emit_reasoning_delta(item_id=item_id, text=final_text)
        self.emitter.emit_reasoning_ended(item_id=item_id)

    def __call__(self, event: dict[str, Any]) -> None:
        event_type = _as_str(event.get("type"))
        if event_type is None:
            return

        if event_type in ("agent.event", "codex.event"):
            if self.custom_event_callback is not None:
                self.custom_event_callback(event)
            return

        if event_type.startswith("response."):
            self._track_response_boundary(event=event)

        if event_type == "response.output_item.added":
            self._on_output_item(event=event, completed=False)
            return

        if event_type == "response.output_item.done":
            self._on_output_item(event=event, completed=True)
            return

        if event_type == "response.content_part.added":
            self._on_content_part_added(event=event)
            return

        if event_type == "response.content_part.done":
            self._on_content_part_done(event=event)
            return

        if event_type in {"response.output_text.delta", "response.refusal.delta"}:
            self._on_text_delta(event=event)
            return

        if event_type == "response.output_text.done":
            self._on_text_done(event=event, field_name="text")
            return

        if event_type == "response.refusal.done":
            self._on_text_done(event=event, field_name="refusal")
            return

        if event_type in {
            "response.reasoning_text.delta",
            "response.reasoning_summary_text.delta",
        }:
            self._on_reasoning_delta(event=event, field_name="delta")
            return

        if event_type == "response.reasoning_summary_part.added":
            part = _as_dict(event.get("part"))
            item_id = self._item_id_from_event(event=event)
            if part is None or item_id is None:
                return
            self.emitter.emit_reasoning_started(item_id=item_id)
            self.emitter.emit_reasoning_delta(item_id=item_id, text=_part_text(part))
            return

        if event_type == "response.reasoning_summary_part.done":
            self._on_reasoning_done(event=event)
            return

        if event_type == "response.reasoning_text.done":
            self._on_reasoning_done(event=event, field_name="text")
            return

        if event_type == "response.reasoning_summary_text.done":
            self._on_reasoning_done(event=event)
            return

        if event_type == "meshagent.handler.added":
            item = _as_dict(event.get("item"))
            if item is None:
                return
            tool_info = _openai_tool_call_info(
                item,
                function_tool_name_resolver=self.function_tool_name_resolver,
            )
            if tool_info is None:
                return
            self._pending_handler_tool_calls[tool_info.item_id] = tool_info
            self.emitter.emit_tool_started(info=tool_info)
            return

        if event_type == "meshagent.handler.output":
            item_id = _as_str(event.get("item_id"))
            if item_id is None:
                return
            self.emitter.emit_tool_log_delta(
                item_id=item_id,
                lines=_parse_tool_log_lines(event.get("lines")),
            )
            return

        if event_type == "meshagent.handler.done":
            result_item = _as_dict(event.get("item"))
            item_id = _as_str(event.get("item_id"))
            if item_id is None and result_item is not None:
                item_id = _as_str(result_item.get("id"))

            pending_tool_call = (
                self._pending_handler_tool_calls.pop(item_id, None)
                if item_id is not None
                else None
            )
            if pending_tool_call is None and result_item is not None:
                result_call_id = _as_str(result_item.get("call_id"))
                if result_call_id is not None:
                    for pending_item_id, pending_info in list(
                        self._pending_handler_tool_calls.items()
                    ):
                        if pending_info.call_id != result_call_id:
                            continue
                        pending_tool_call = pending_info
                        self._pending_handler_tool_calls.pop(pending_item_id, None)
                        break
            if pending_tool_call is None and item_id is None:
                if len(self._pending_handler_tool_calls) == 1:
                    pending_item_id, pending_info = next(
                        iter(self._pending_handler_tool_calls.items())
                    )
                    pending_tool_call = pending_info
                    self._pending_handler_tool_calls.pop(pending_item_id, None)
            if pending_tool_call is None and result_item is not None:
                pending_tool_call = _openai_tool_call_info(
                    result_item,
                    function_tool_name_resolver=self.function_tool_name_resolver,
                )
            if pending_tool_call is None:
                return

            result = pending_tool_call.result
            if result_item is not None:
                result = (
                    _content_from_tool_item(result_item)
                    or _content_from_handler_result(result_item.get("result"))
                    or result
                )
            else:
                result = (
                    _content_from_handler_result(event.get("result"))
                    or pending_tool_call.result
                )

            error_text = _as_str(event.get("error"))
            self._finished_handler_tool_call_ids.add(pending_tool_call.item_id)
            self.emitter.emit_tool_ended(
                info=replace(
                    pending_tool_call,
                    result=result,
                    error=(
                        AgentError(
                            message=error_text,
                            code="tool_call_failed",
                        )
                        if error_text is not None
                        else pending_tool_call.error
                    ),
                )
            )


@dataclass(slots=True)
class _AnthropicBlockState:
    kind: Literal["file", "reasoning", "text", "tool"]
    item_id: str
    block: dict[str, Any]
    arguments_text: str = ""


@dataclass(slots=True)
class _AnthropicAgentEventPublisher:
    emitter: _AgentMessageEmitter
    function_tool_name_resolver: FunctionToolNameResolver | None = None
    custom_event_callback: Callable[[dict[str, Any]], None] | None = None
    _openai_publisher: _OpenAIAgentEventPublisher = field(init=False)
    _message_id: str | None = None
    _blocks: dict[int, _AnthropicBlockState] = field(default_factory=dict)

    def __post_init__(self) -> None:
        self._openai_publisher = _OpenAIAgentEventPublisher(
            emitter=self.emitter,
            function_tool_name_resolver=self.function_tool_name_resolver,
            custom_event_callback=self.custom_event_callback,
        )

    def set_function_tool_name_resolver(
        self,
        resolver: FunctionToolNameResolver | None,
    ) -> None:
        self.function_tool_name_resolver = resolver
        self._openai_publisher.set_function_tool_name_resolver(resolver)

    def _synthesized_item_id(self, *, index: int) -> str:
        base = self._message_id or "anthropic"
        return f"{base}:content:{index}"

    def __call__(self, event: dict[str, Any]) -> None:
        event_type = _as_str(event.get("type"))
        if event_type is None:
            return

        if event_type in ("agent.event", "codex.event"):
            if self.custom_event_callback is not None:
                self.custom_event_callback(event)
            return

        if event_type.startswith("response."):
            self._openai_publisher(event)
            return

        if event_type.startswith("meshagent.handler."):
            self._openai_publisher(event)
            return

        normalized_type = event_type.replace(".", "_")
        payload = _as_dict(event.get("event")) or {}

        if normalized_type == "message_start":
            message = _as_dict(payload.get("message"))
            if message is None:
                return
            self._message_id = _as_str(message.get("id"))
            return

        if normalized_type == "content_block_start":
            index = _as_int(payload.get("index"))
            block = _as_dict(payload.get("content_block"))
            if index is None or block is None:
                return

            block_type = _as_str(block.get("type"))
            if block_type is None:
                return

            if block_type == "text":
                item_id = self._synthesized_item_id(index=index)
                self._blocks[index] = _AnthropicBlockState(
                    kind="text",
                    item_id=item_id,
                    block=block,
                )
                self.emitter.emit_text_started(item_id=item_id)
                self.emitter.emit_text_delta(
                    item_id=item_id,
                    text=_as_text(block.get("text")) or "",
                )
                return

            if block_type == "thinking":
                item_id = self._synthesized_item_id(index=index)
                self._blocks[index] = _AnthropicBlockState(
                    kind="reasoning",
                    item_id=item_id,
                    block=block,
                )
                self.emitter.emit_reasoning_started(item_id=item_id)
                self.emitter.emit_reasoning_delta(
                    item_id=item_id,
                    text=_as_text(block.get("thinking")) or "",
                )
                return

            tool_item_id = _as_str(block.get("id"))
            tool_info = (
                _anthropic_tool_call_info(
                    block=block,
                    item_id=tool_item_id,
                    function_tool_name_resolver=self.function_tool_name_resolver,
                )
                if tool_item_id is not None
                else None
            )
            if tool_info is not None:
                self._blocks[index] = _AnthropicBlockState(
                    kind="tool",
                    item_id=tool_info.item_id,
                    block=block,
                )
                self.emitter.emit_tool_pending(info=tool_info)
                return

            if block_type in {"document", "image"}:
                item_id = self._synthesized_item_id(index=index)
                self._blocks[index] = _AnthropicBlockState(
                    kind="file",
                    item_id=item_id,
                    block=block,
                )
                self.emitter.emit_file_started(item_id=item_id)
                self.emitter.emit_file_delta(
                    item_id=item_id,
                    url=_file_url_from_payload(payload=block, provider="anthropic"),
                )
            return

        if normalized_type == "content_block_delta":
            index = _as_int(payload.get("index"))
            delta = _as_dict(payload.get("delta"))
            if index is None or delta is None:
                return

            state = self._blocks.get(index)
            if state is None:
                return

            delta_type = _as_str(delta.get("type"))
            if state.kind == "text" and delta_type == "text_delta":
                self.emitter.emit_text_delta(
                    item_id=state.item_id,
                    text=_as_text(delta.get("text")) or "",
                )
                return

            if state.kind == "reasoning" and delta_type == "thinking_delta":
                self.emitter.emit_reasoning_delta(
                    item_id=state.item_id,
                    text=_as_text(delta.get("thinking")) or "",
                )
                return

            if state.kind == "tool" and delta_type == "input_json_delta":
                state.arguments_text += _as_text(delta.get("partial_json")) or ""
                return

            if state.kind == "file":
                self.emitter.emit_file_delta(
                    item_id=state.item_id,
                    url=_file_url_from_payload(payload=delta, provider="anthropic"),
                )
            return

        if normalized_type == "content_block_stop":
            index = _as_int(payload.get("index"))
            if index is None:
                return

            state = self._blocks.pop(index, None)
            if state is None:
                return

            if state.kind == "text":
                self.emitter.emit_text_ended(item_id=state.item_id)
                return

            if state.kind == "reasoning":
                self.emitter.emit_reasoning_ended(item_id=state.item_id)
                return

            if state.kind == "file":
                if not self.emitter.has_content_data(
                    kind="file", item_id=state.item_id
                ):
                    self.emitter.emit_file_delta(
                        item_id=state.item_id,
                        url=_file_url_from_payload(
                            payload=state.block,
                            provider="anthropic",
                        ),
                    )
                self.emitter.emit_file_ended(item_id=state.item_id)
                return

            tool_info = _anthropic_tool_call_info(
                block=state.block,
                item_id=state.item_id,
                function_tool_name_resolver=self.function_tool_name_resolver,
            )
            if tool_info is not None:
                parsed_arguments = (
                    _parse_tool_arguments(state.arguments_text)
                    if state.arguments_text != ""
                    else None
                )
                updated_arguments = parsed_arguments
                if updated_arguments is None:
                    updated_arguments = tool_info.arguments
                self.emitter.emit_tool_pending(
                    info=replace(tool_info, arguments=updated_arguments)
                )


def make_openai_agent_event_publisher(
    *,
    turn_id: str,
    thread_id: str,
    callback: AgentEventCallback,
    function_tool_name_resolver: FunctionToolNameResolver | None = None,
    custom_event_callback: Callable[[dict[str, Any]], None] | None = None,
) -> Callable[[dict[str, Any]], None]:
    emitter = _AgentMessageEmitter(
        turn_id=turn_id,
        thread_id=thread_id,
        callback=callback,
    )
    publisher = _OpenAIAgentEventPublisher(
        emitter=emitter,
        function_tool_name_resolver=function_tool_name_resolver,
        custom_event_callback=custom_event_callback,
    )
    return publisher


def make_anthropic_agent_event_publisher(
    *,
    turn_id: str,
    thread_id: str,
    callback: AgentEventCallback,
    function_tool_name_resolver: FunctionToolNameResolver | None = None,
    custom_event_callback: Callable[[dict[str, Any]], None] | None = None,
) -> Callable[[dict[str, Any]], None]:
    emitter = _AgentMessageEmitter(
        turn_id=turn_id,
        thread_id=thread_id,
        callback=callback,
    )
    publisher = _AnthropicAgentEventPublisher(
        emitter=emitter,
        function_tool_name_resolver=function_tool_name_resolver,
        custom_event_callback=custom_event_callback,
    )
    return publisher



import json
from abc import ABC, abstractmethod
from collections.abc import Awaitable, Callable
from dataclasses import dataclass
from typing import Any, Generic, Literal, Optional, TypeVar

from jsonschema import validate

from meshagent.api.messaging import FileContent, JsonContent, TextContent
from meshagent.api import RoomClient, RoomException, RemoteParticipant
from meshagent.tools import Content, ToolContext, Toolkit, ToolkitBuilder, ToolkitConfig

from .agent import AgentSessionContext
from .messages import AgentMessage

TEvent = TypeVar("T")

DEFAULT_MAX_TOOL_CALL_LINES = 2000
DEFAULT_MAX_TOOL_CALL_LENGTH = 50 * 1024
_TEXTUAL_APPLICATION_MIME_TYPES = {
    "application/json",
    "application/javascript",
    "application/xml",
    "application/x-sh",
    "application/yaml",
    "application/x-yaml",
    "application/xhtml+xml",
}


@dataclass(frozen=True, slots=True)
class ToolCallApprovalRequest:
    item_id: str
    toolkit: str
    tool: str
    arguments: dict[str, Any] | None = None


ToolCallApprovalHandler = Callable[
    [ToolContext, ToolCallApprovalRequest],
    Awaitable[bool],
]
SteeringCallback = Callable[[], Awaitable[bool]]


class ToolResponseAdapter(ABC):
    def __init__(
        self,
        *,
        max_tool_call_length: int = DEFAULT_MAX_TOOL_CALL_LENGTH,
        max_tool_call_lines: int = DEFAULT_MAX_TOOL_CALL_LINES,
    ):
        if max_tool_call_length <= 0:
            raise ValueError("max_tool_call_length must be greater than 0")
        if max_tool_call_lines <= 0:
            raise ValueError("max_tool_call_lines must be greater than 0")
        self.max_tool_call_length = max_tool_call_length
        self.max_tool_call_lines = max_tool_call_lines

    @staticmethod
    def _normalize_mime_type(mime_type: str | None) -> str:
        if mime_type is None:
            return ""
        return mime_type.partition(";")[0].strip().lower()

    @staticmethod
    def _looks_like_text(*, data: bytes, decoded: str) -> bool:
        if b"\x00" in data:
            return False
        return all(ord(ch) >= 32 or ch in "\n\r\t\f\b" for ch in decoded)

    async def file_content_to_text_content(
        self,
        *,
        room: RoomClient,
        content: Content,
    ) -> TextContent | None:
        del room
        if not isinstance(content, FileContent):
            return None

        normalized_mime_type = self._normalize_mime_type(content.mime_type)
        is_declared_text = normalized_mime_type.startswith("text/") or (
            normalized_mime_type in _TEXTUAL_APPLICATION_MIME_TYPES
        )

        try:
            decoded = content.data.decode("utf-8")
        except UnicodeDecodeError:
            return None

        if is_declared_text or self._looks_like_text(
            data=content.data,
            decoded=decoded,
        ):
            return TextContent(text=decoded)

        return None

    def truncate(self, *, room: RoomClient, content: Content) -> Content:
        del room
        text: str | None = None
        if isinstance(content, TextContent):
            text = content.text
        elif isinstance(content, JsonContent):
            text = json.dumps(content.json, ensure_ascii=False)

        if text is None:
            return content

        original_line_count = len(text.splitlines()) if text != "" else 0
        limited_text = text
        if original_line_count > self.max_tool_call_lines:
            limited_text = "\n".join(text.splitlines()[: self.max_tool_call_lines])

        original_bytes = text.encode("utf-8")
        limited_bytes = limited_text.encode("utf-8")
        if len(limited_bytes) > self.max_tool_call_length:
            limited_text = limited_bytes[: self.max_tool_call_length].decode(
                "utf-8", errors="ignore"
            )

        if (
            original_line_count <= self.max_tool_call_lines
            and len(original_bytes) <= self.max_tool_call_length
        ):
            return content

        truncated_text = limited_text.rstrip()
        notice = (
            "The tool call returned too much data and was truncated. "
            f"Showing at most {self.max_tool_call_lines} lines and "
            f"{self.max_tool_call_length} bytes."
        )
        if truncated_text == "":
            return TextContent(text=notice)

        return TextContent(text=f"{truncated_text}\n\n{notice}")

    @abstractmethod
    async def to_plain_text(self, *, room: RoomClient, response: Content):
        pass

    @abstractmethod
    async def create_messages(
        self,
        *,
        context: AgentSessionContext,
        tool_call: Any,
        room: RoomClient,
        response: Content,
    ) -> list:
        pass


class LLMAdapter(Generic[TEvent]):
    outputTokenMax: float = float("inf")

    @abstractmethod
    def default_model(self) -> str: ...

    def create_session(self) -> AgentSessionContext:
        return AgentSessionContext()

    def get_additional_instructions(self) -> str | None:
        return None

    def on_turn_steer(self, *, context: AgentSessionContext, interrupted: bool) -> None:
        del context
        del interrupted

    def context_window_size(self, model: str) -> float:
        return float("inf")

    def needs_compaction(self, *, context: AgentSessionContext) -> bool:
        return False

    async def compact(
        self,
        *,
        context: AgentSessionContext,
        room: RoomClient,
        model: Optional[str] = None,
    ) -> None:
        return None

    async def get_input_tokens(
        self,
        *,
        context: AgentSessionContext,
        model: str,
        room: Optional[RoomClient] = None,
        toolkits: Optional[list] = None,
        output_schema: Optional[dict] = None,
    ) -> int:
        return 0

    async def check_for_termination(
        self, *, context: AgentSessionContext, room: RoomClient
    ):
        return True

    def set_tool_call_approval_handler(
        self, handler: ToolCallApprovalHandler | None
    ) -> None:
        del handler

    def make_agent_event_publisher(
        self,
        turn_id: str,
        thread_id: str,
        callback: Callable[[AgentMessage], None],
        custom_event_callback: Callable[[dict[str, Any]], None] | None = None,
    ) -> Callable[[TEvent], None]:
        del turn_id
        del thread_id

        def publish(event: TEvent) -> None:
            if isinstance(event, AgentMessage):
                callback(event)
                return

            if not isinstance(event, dict) or custom_event_callback is None:
                return

            event_type = event.get("type")
            if event_type in ("agent.event", "codex.event"):
                custom_event_callback(event)

        return publish

    def tool_providers(self, *, model: str) -> list[ToolkitBuilder]:
        return []

    async def make_toolkit(
        self, *, room: RoomClient, model: str, config: ToolkitConfig
    ) -> Toolkit:
        for tool in self.tool_providers(model=model):
            if tool.name == config.name:
                return Toolkit(
                    name=config.name,
                    tools=[await tool.make(room=room, model=model, config=config)],
                )

        raise RoomException(f"Unexpected tool: {config.name} for model {model}")

    @abstractmethod
    async def next(
        self,
        *,
        context: AgentSessionContext,
        room: RoomClient,
        toolkits: list[Toolkit],
        output_schema: Optional[dict] = None,
        event_handler: Optional[Callable[[TEvent], None]] = None,
        steering_callback: SteeringCallback | None = None,
        model: Optional[str] = None,
        on_behalf_of: Optional[RemoteParticipant] = None,
        options: Optional[dict] = None,
    ) -> Any:
        pass

    def validate(response: dict, output_schema: dict):
        validate(response, output_schema)


class MessageStreamLLMAdapter(LLMAdapter):
    def __init__(
        self, *, participant_name: str, context_mode: Literal["diff", "full"] = "diff"
    ):
        self.participant_name = participant_name
        self.context_mode = context_mode

    def default_model(self) -> str:
        return "toolkit"

    def create_session(self) -> AgentSessionContext:
        return AgentSessionContext()

    async def check_for_termination(
        self, *, context: AgentSessionContext, room: RoomClient
    ):
        return True

    async def next(
        self,
        *,
        context: AgentSessionContext,
        room: RoomClient,
        toolkits: list[Toolkit],
        output_schema: Optional[dict] = None,
        event_handler: Optional[Callable[[TEvent], None]] = None,
        steering_callback: SteeringCallback | None = None,
        model: Optional[str] = None,
        on_behalf_of: Optional[RemoteParticipant] = None,
        options: Optional[dict] = None,
    ) -> Any:
        del context
        del room
        del toolkits
        del output_schema
        del event_handler
        del steering_callback
        del model
        del on_behalf_of
        del options
        raise RoomException(
            "MessageStreamLLMAdapter has been removed; use streaming toolkits instead"
        )

    def validate(response: dict, output_schema: dict):
        validate(response, output_schema)



from __future__ import annotations

import pytest

from .shell_semantics import analyze_shell_command


def test_analyze_shell_command_coalesces_cd_prefixed_exploration_chain() -> None:
    analysis = analyze_shell_command(
        command=(
            "cd /website && pwd && ls -la && find . -maxdepth 2 -type f | "
            "sed 's#^./##' | sort | head -200"
        )
    )

    assert [op.kind for op in analysis.operations] == ["explore"]
    assert analysis.operations[0].mode == "explore"
    assert analysis.operations[0].path == "/website"
    assert analysis.display.event_kind == "exec"
    assert analysis.display.path == "/website"
    assert analysis.display.coalesce_path == "/website"
    assert analysis.display.phase_for_state(state="pending").headline == (
        "Preparing to explore /website"
    )
    assert analysis.display.phase_for_state(state="in_progress").headline == (
        "Exploring /website"
    )
    assert analysis.display.phase_for_state(state="completed").headline == (
        "Explored /website"
    )


def test_analyze_shell_command_coalesces_cd_prefixed_read_chain() -> None:
    analysis = analyze_shell_command(
        command=(
            "cd /website && sed -n '1,220p' public/index.html && "
            "printf '\\n---CSS---\\n' && sed -n '1,260p' public/styles.css && "
            "printf '\\n---JS---\\n' && sed -n '1,260p' public/app.js"
        )
    )

    assert [op.kind for op in analysis.operations] == ["explore"]
    assert analysis.operations[0].path == "/website"
    assert analysis.display.phase_for_state(state="pending").headline == (
        "Preparing to explore /website"
    )
    assert analysis.display.phase_for_state(state="in_progress").headline == (
        "Exploring /website"
    )


def test_analyze_shell_command_renders_single_heredoc_write() -> None:
    analysis = analyze_shell_command(
        command=(
            "cd /website && cat > public/index.html <<'EOF'\n"
            "<!doctype html>\n"
            "<html></html>\n"
            "EOF"
        )
    )

    assert [op.kind for op in analysis.operations] == ["write"]
    assert analysis.operations[0].paths == ("/website/public/index.html",)
    assert analysis.display.event_kind == "file"
    assert analysis.display.path == "/website/public/index.html"
    assert analysis.display.preview == (
        "cd /website && cat > public/index.html <<'EOF'\n"
        "<!doctype html>\n"
        "<html></html>\n"
        "EOF"
    )
    assert analysis.display.phase_for_state(state="pending").headline == (
        "Preparing to write /website/public/index.html"
    )
    assert analysis.display.phase_for_state(state="in_progress").headline == (
        "Writing /website/public/index.html"
    )
    assert analysis.display.phase_for_state(state="completed").headline == (
        "Wrote /website/public/index.html"
    )
    assert analysis.display.phase_for_state(state="failed").headline == (
        "Attempted to write file /website/public/index.html"
    )


def test_analyze_shell_command_preview_truncates_after_five_lines() -> None:
    analysis = analyze_shell_command(
        command=(
            "cd /website && cat > public/index.html <<'EOF'\n"
            "line 1\n"
            "line 2\n"
            "line 3\n"
            "line 4\n"
            "line 5\n"
            "line 6\n"
            "EOF"
        )
    )

    assert analysis.display.preview == (
        "cd /website && cat > public/index.html <<'EOF'\n"
        "line 1\n"
        "line 2\n"
        "line 3\n"
        "line 4\n"
        "..."
    )


def test_analyze_shell_command_groups_multi_file_heredoc_writes() -> None:
    analysis = analyze_shell_command(
        command=(
            "cd /website && mkdir -p src public dist && "
            "cat > package.json <<'EOF'\n{}\nEOF\n"
            "cat > tsconfig.json <<'EOF'\n{}\nEOF\n"
            "cat > src/main.tsx <<'EOF'\nconsole.log('hi')\nEOF\n"
            "npm install\n"
            "npm run build"
        )
    )

    assert analysis.operations[0].kind == "write"
    assert analysis.operations[0].path == "/website"
    assert analysis.operations[0].paths == (
        "/website/package.json",
        "/website/tsconfig.json",
        "/website/src/main.tsx",
    )
    assert analysis.display.event_kind == "file"
    assert analysis.display.preview == (
        "cd /website && mkdir -p src public dist && cat > package.json <<'EOF'\n"
        "{}\n"
        "EOF\n"
        "cat > tsconfig.json <<'EOF'\n"
        "{}\n"
        "..."
    )
    assert analysis.display.phase_for_state(state="pending").headline == (
        "Preparing to write files in /website"
    )
    assert analysis.display.phase_for_state(state="in_progress").headline == (
        "Writing files in /website"
    )
    assert analysis.display.phase_for_state(state="completed").headline == (
        "Wrote files in /website"
    )


def test_analyze_shell_command_treats_python_heredoc_file_generation_as_write() -> None:
    analysis = analyze_shell_command(
        command=(
            "cd /data/docs && python - <<'PY'\n"
            "for name in ['index.html', 'faq.html']:\n"
            "    with open(name, 'w', encoding='utf-8') as f:\n"
            "        f.write('hello')\n"
            "PY\n"
            "ls -1 /data/docs"
        )
    )

    assert analysis.operations[0].kind == "write"
    assert analysis.operations[0].path == "/data/docs"
    assert analysis.operations[0].multi is True
    assert analysis.operations[-1].kind == "explore"
    assert analysis.display.event_kind == "file"
    assert analysis.display.phase_for_state(state="pending").headline == (
        "Preparing to write files in /data/docs"
    )
    assert analysis.display.phase_for_state(state="in_progress").headline == (
        "Writing files in /data/docs"
    )
    assert analysis.display.phase_for_state(state="completed").headline == (
        "Wrote files in /data/docs"
    )


def test_analyze_shell_command_treats_if_guarded_heredoc_write_as_write() -> None:
    analysis = analyze_shell_command(
        command=(
            "mkdir -p /website/docs && cd /website/docs && "
            "if [ ! -f index.html ]; then cat > index.html <<'EOF'\n"
            "<!doctype html>\n"
            "EOF\n"
            "fi"
        )
    )

    assert analysis.operations[0].kind == "write"
    assert analysis.operations[0].path == "/website/docs/index.html"
    assert analysis.operations[0].paths == ("/website/docs/index.html",)
    assert analysis.display.event_kind == "file"
    assert analysis.display.path == "/website/docs/index.html"
    assert analysis.display.phase_for_state(state="pending").headline == (
        "Preparing to write /website/docs/index.html"
    )
    assert analysis.display.phase_for_state(state="in_progress").headline == (
        "Writing /website/docs/index.html"
    )
    assert analysis.display.phase_for_state(state="completed").headline == (
        "Wrote /website/docs/index.html"
    )


def test_analyze_shell_command_treats_for_loop_grep_as_search_in_cwd() -> None:
    analysis = analyze_shell_command(
        command=(
            "cd /tmp/meshsite && for f in home.html meshagent-sdk.html "
            "meshagent-server.html meshagent-studio.html powerboards.html "
            "faqs.html; do echo '===== '$f; grep -oiE "
            "'agent|sdk|server|studio|powerboards|room|api|monitor|orchestrat|"
            "deploy|runtime|workflow|realtime|tool|memory|observ|trace|mcp|"
            'integration|enterprise\' "$f" | sort | uniq -c | sort -nr | '
            "head -40; done"
        )
    )

    assert [op.kind for op in analysis.operations] == ["explore", "search"]
    assert analysis.operations[-1].path == "/tmp/meshsite"
    assert analysis.display.phase_for_state(state="pending").headline == (
        "Preparing to search /tmp/meshsite"
    )
    assert analysis.display.phase_for_state(state="in_progress").headline == (
        "Searching /tmp/meshsite"
    )
    assert analysis.display.phase_for_state(state="completed").headline == (
        "Searched /tmp/meshsite"
    )


def test_analyze_shell_command_falls_back_to_shell_script_for_complex_parse_failures() -> (
    None
):
    analysis = analyze_shell_command(
        command="cd /tmp/meshsite && for f in a b; do echo $f"
    )

    assert [op.kind for op in analysis.operations] == ["script"]
    assert analysis.operations[0].path == "/tmp/meshsite"
    assert analysis.display.phase_for_state(state="pending").headline == (
        "Preparing to run shell script in /tmp/meshsite"
    )
    assert analysis.display.phase_for_state(state="in_progress").headline == (
        "Running shell script in /tmp/meshsite"
    )
    assert analysis.display.phase_for_state(state="completed").headline == (
        "Ran shell script in /tmp/meshsite"
    )


def test_analyze_shell_command_tracks_install_build_and_exploration_sequence() -> None:
    analysis = analyze_shell_command(
        command=(
            "cd /website && npm install tslib @types/react @types/react-dom "
            "--cache /agents/webmaster/.npm && npm run build && "
            "find dist -maxdepth 2 -type f | sort"
        )
    )

    assert [op.kind for op in analysis.operations] == ["install", "build", "explore"]
    assert analysis.operations[0].path == "/website"
    assert analysis.operations[1].path == "/website"
    assert analysis.operations[2].path == "/website/dist"
    assert analysis.display.phase_for_state(state="pending").headline == (
        "Preparing to install packages and preparing to build project in /website"
    )
    assert analysis.display.phase_for_state(state="in_progress").headline == (
        "Installing packages and building project in /website"
    )
    assert analysis.display.phase_for_state(state="completed").headline == (
        "Installed packages and built project in /website"
    )


def test_analyze_shell_command_combines_install_and_lint_in_same_directory() -> None:
    analysis = analyze_shell_command(
        command=(
            "cd /data/pythontest && "
            "python3 -m pip install ruff >/tmp/ruff_install.log 2>&1 && "
            "python3 -m ruff check ."
        )
    )

    assert [op.kind for op in analysis.operations] == ["install", "lint"]
    assert analysis.display.phase_for_state(state="pending").headline == (
        "Preparing to install packages and preparing to check code in /data/pythontest"
    )
    assert analysis.display.phase_for_state(state="in_progress").headline == (
        "Installing packages and checking code in /data/pythontest"
    )
    assert analysis.display.phase_for_state(state="completed").headline == (
        "Installed packages and checked code in /data/pythontest"
    )
    assert analysis.display.preview == (
        "cd /data/pythontest && python3 -m pip install ruff >/tmp/ruff_install.log 2>&1 && python3 -m ruff check ."
    )


def test_analyze_shell_command_tracks_edit_then_build_sequence() -> None:
    analysis = analyze_shell_command(
        command=(
            "cd /website && stat -c '%A %n' node_modules/rollup/dist/bin/rollup "
            "node_modules/.bin/rollup && sed -i "
            '\'s#"build": "rollup -c"#"build": '
            '"node ./node_modules/rollup/dist/bin/rollup -c"#\' package.json && '
            'sed -i \'s#"dev": "rollup -c -w"#"dev": '
            '"node ./node_modules/rollup/dist/bin/rollup -c -w"#\' package.json && '
            "npm run build"
        )
    )

    assert [op.kind for op in analysis.operations] == [
        "explore",
        "edit",
        "edit",
        "build",
    ]
    assert analysis.operations[0].path == "/website/node_modules"
    assert analysis.operations[1].path == "/website/package.json"
    assert analysis.operations[2].path == "/website/package.json"
    assert analysis.operations[3].path == "/website"
    assert analysis.display.phase_for_state(state="in_progress").headline == (
        "Building /website"
    )


def test_analyze_shell_command_recognizes_search_display() -> None:
    analysis = analyze_shell_command(
        command="cd /website && rg --ignore-case hero public/index.html"
    )

    assert [op.kind for op in analysis.operations] == ["search"]
    assert analysis.operations[0].query == "hero"
    assert analysis.operations[0].path == "/website/public/index.html"
    assert analysis.display.path == "/website/public/index.html"
    assert analysis.display.coalesce_path == "/website/public/index.html"
    assert analysis.display.phase_for_state(state="pending").headline == (
        "Preparing to search /website/public/index.html"
    )
    assert analysis.display.phase_for_state(state="in_progress").headline == (
        "Searching /website/public/index.html"
    )
    assert analysis.display.details == ("Pattern: hero",)


def test_analyze_shell_command_recognizes_download_display() -> None:
    analysis = analyze_shell_command(
        command="cd /website && curl -L https://example.com/logo.svg -o public/logo.svg"
    )

    assert [op.kind for op in analysis.operations] == ["download"]
    assert analysis.operations[0].path == "/website/public/logo.svg"
    assert analysis.display.event_kind == "file"
    assert analysis.display.phase_for_state(state="pending").headline == (
        "Preparing to download /website/public/logo.svg"
    )
    assert analysis.display.phase_for_state(state="in_progress").headline == (
        "Downloading /website/public/logo.svg"
    )
    assert analysis.display.phase_for_state(state="completed").headline == (
        "Downloaded /website/public/logo.svg"
    )
    assert analysis.display.preview == (
        "cd /website && curl -L https://example.com/logo.svg -o public/logo.svg"
    )


def test_analyze_shell_command_treats_curl_without_output_as_web_request() -> None:
    analysis = analyze_shell_command(command="curl -sS http://127.0.0.1:8080/")

    assert [op.kind for op in analysis.operations] == ["request"]
    assert analysis.display.phase_for_state(state="pending").headline == (
        "Preparing web request"
    )
    assert analysis.display.phase_for_state(state="in_progress").headline == (
        "Making web request"
    )
    assert analysis.display.preview == "curl -sS http://127.0.0.1:8080/"
    assert analysis.display.details == ("http://127.0.0.1:8080/",)


def test_analyze_shell_command_falls_back_to_run_display() -> None:
    analysis = analyze_shell_command(command="node scripts/custom.js --flag")

    assert [op.kind for op in analysis.operations] == ["run"]
    assert analysis.display.event_kind == "exec"
    assert analysis.display.path == ""
    assert analysis.display.phase_for_state(state="pending").headline == (
        "Preparing Command"
    )
    assert analysis.display.phase_for_state(state="in_progress").headline == (
        "Running Command"
    )
    assert analysis.display.phase_for_state(state="failed").headline == (
        "Attempted to run command"
    )
    assert analysis.display.preview == "node scripts/custom.js --flag"


@pytest.mark.parametrize(
    ("command", "expected_kind", "expected_headline"),
    [
        ("npx vite build", "build", "Building project"),
        ("bunx vite build", "build", "Building project"),
        ("astro build", "build", "Building project"),
        (
            "npm exec webpack -- --mode production",
            "build",
            "Building project",
        ),
        (
            "node ./node_modules/rollup/dist/bin/rollup -c",
            "build",
            "Building project",
        ),
        ("npm run dev", "dev", "Starting dev command in project"),
        ("next dev", "dev", "Starting dev command in project"),
        ("webpack serve", "dev", "Starting dev command in project"),
        ("uv run pytest -q", "test", "Running tests in project"),
        ("python -m pytest -q", "test", "Running tests in project"),
        ("pnpm exec vitest run", "test", "Running tests in project"),
        ("npx playwright test", "test", "Running tests in project"),
        ("cargo test", "test", "Running tests in project"),
        ("uv run ruff check .", "lint", "Checking code in project"),
        ("npx prettier --write .", "lint", "Checking code in project"),
        ("pnpm dlx prettier --write .", "lint", "Checking code in project"),
        ("cargo clippy", "lint", "Checking code in project"),
        ("poetry add requests", "install", "Installing packages in project"),
        ("composer install", "install", "Installing packages in project"),
        ("git grep hero src", "search", "Searching src"),
        ("git status --short", "explore", "Exploring project"),
        ("fd main src", "explore", "Exploring src"),
    ],
)
def test_analyze_shell_command_supports_common_dev_clis_and_utilities(
    *,
    command: str,
    expected_kind: str,
    expected_headline: str,
) -> None:
    analysis = analyze_shell_command(command=command)

    assert analysis.operations[0].kind == expected_kind
    assert (
        analysis.display.phase_for_state(state="in_progress").headline
        == expected_headline
    )



from abc import ABC, abstractmethod
from typing import Optional

from meshagent.api import RoomClient, RoomException

from .adapter import LLMAdapter

_DEFAULT_IMAGE_CAPTION_SCHEMA = {
    "type": "object",
    "additionalProperties": False,
    "required": ["caption"],
    "properties": {
        "caption": {
            "type": "string",
            "description": "a concise human-readable caption for the image",
        }
    },
}

_DEFAULT_IMAGE_CAPTION_PROMPT = (
    "Write a concise caption for this image. Focus on the important visual content."
)


class ImageCaptioner(ABC):
    @abstractmethod
    async def caption(
        self,
        *,
        room: RoomClient,
        image_data: bytes,
        mime_type: str,
    ) -> str: ...


class LLMImageCaptioner(ImageCaptioner):
    def __init__(
        self,
        *,
        llm_adapter: LLMAdapter,
        output_schema: Optional[dict] = None,
        rules: Optional[list[str]] = None,
        prompt: Optional[str] = None,
    ):
        self._llm_adapter = llm_adapter
        self._output_schema = (
            output_schema
            if output_schema is not None
            else _DEFAULT_IMAGE_CAPTION_SCHEMA.copy()
        )
        self._rules = [*(rules or [])]
        self._prompt = (
            prompt
            if isinstance(prompt, str) and prompt.strip() != ""
            else _DEFAULT_IMAGE_CAPTION_PROMPT
        )

    @staticmethod
    def _extract_caption(response: object) -> str:
        if isinstance(response, str):
            return response.strip()

        if isinstance(response, dict):
            for key in ("caption", "text", "description", "summary"):
                value = response.get(key)
                if isinstance(value, str) and value.strip() != "":
                    return value.strip()

        return str(response).strip()

    async def caption(
        self,
        *,
        room: RoomClient,
        image_data: bytes,
        mime_type: str,
    ) -> str:
        session_context = self._llm_adapter.create_session()
        async with session_context:
            if not session_context.supports_images:
                raise RoomException(
                    "llm adapter chat context does not support image inputs for captioning"
                )

            if len(self._rules) > 0:
                session_context.append_rules(self._rules)

            session_context.append_user_message(self._prompt)
            session_context.append_image_message(mime_type=mime_type, data=image_data)

            response = await self._llm_adapter.next(
                context=session_context,
                room=room,
                toolkits=[],
                output_schema=self._output_schema,
            )
            return self._extract_caption(response)



from __future__ import annotations

import logging
from typing import Any

from meshagent.api import RoomMessage
from pydantic import BaseModel
from meshagent.tools import RemoteToolkit

from .legacy_chat_channel import LegacyChatChannel
from .messages import (
    AGENT_EVENT_THREAD_CLEARED,
    AGENT_MESSAGE_THREAD_CLEAR,
    AGENT_MESSAGE_TOOL_CALL_APPROVE,
    AGENT_MESSAGE_TOOL_CALL_REJECT,
    AGENT_MESSAGE_TURN_INTERRUPT,
    AGENT_MESSAGE_TURN_START,
    AGENT_MESSAGE_TURN_STEER,
    AgentMessage,
    ApproveAgentToolCall,
    ClearThread,
    RejectAgentToolCall,
    TurnInterrupt,
    TurnStart,
    TurnSteer,
)
from .process import Message

logger = logging.getLogger("chat-channel")


class _AgentMessageEnvelope(BaseModel):
    payload: dict[str, Any]


_INBOUND_AGENT_MESSAGE_MODELS: dict[str, type[AgentMessage]] = {
    AGENT_MESSAGE_TURN_START: TurnStart,
    AGENT_MESSAGE_TURN_STEER: TurnSteer,
    AGENT_MESSAGE_TURN_INTERRUPT: TurnInterrupt,
    AGENT_MESSAGE_THREAD_CLEAR: ClearThread,
    AGENT_MESSAGE_TOOL_CALL_APPROVE: ApproveAgentToolCall,
    AGENT_MESSAGE_TOOL_CALL_REJECT: RejectAgentToolCall,
}

_OUTBOUND_AGENT_MESSAGE_TYPES = {
    AGENT_EVENT_THREAD_CLEARED,
}


class ChatChannel(LegacyChatChannel):
    def get_exposed_toolkits(self) -> list[RemoteToolkit]:
        return [self.make_remote_toolkit()]

    def handles(self, message: Message) -> bool:
        message_type = message.data.type
        if message_type in _INBOUND_AGENT_MESSAGE_MODELS:
            return False
        if message_type in _OUTBOUND_AGENT_MESSAGE_TYPES:
            return True
        return message_type.startswith("meshagent.agent.")

    async def on_start(self) -> None:
        await super().on_start()
        await self.room.local_participant.set_attribute("supports_agent_messages", True)

    async def on_stop(self) -> None:
        await self.room.local_participant.set_attribute("supports_agent_messages", None)
        await super().on_stop()

    async def on_message(self, message: Message) -> None:
        payload = message.data.model_dump(mode="json")
        for participant in self._open_participants(thread_id=message.data.thread_id):
            if participant.id == self.room.local_participant.id:
                continue

            self.room.messaging.send_message_nowait(
                to=participant,
                type="agent-message",
                message={"payload": payload},
            )

    @staticmethod
    def _normalize_agent_message_payload(
        *,
        payload: dict[str, Any],
    ) -> dict[str, Any]:
        normalized_payload = dict(payload)
        message_type = normalized_payload.get("type")
        if message_type not in {
            AGENT_MESSAGE_TURN_START,
            AGENT_MESSAGE_TURN_STEER,
        }:
            return normalized_payload

        raw_content = normalized_payload.get("content")
        if not isinstance(raw_content, list):
            return normalized_payload

        normalized_content: list[Any] = []
        content_changed = False
        for item in raw_content:
            if not isinstance(item, dict):
                normalized_content.append(item)
                continue

            if item.get("type") != "file":
                normalized_content.append(item)
                continue

            url = item.get("url")
            if not isinstance(url, str):
                normalized_content.append(item)
                continue

            normalized_url = LegacyChatChannel._normalize_attachment_url(path=url)
            if normalized_url is None:
                normalized_content.append(item)
                continue

            if normalized_url == url:
                normalized_content.append(item)
                continue

            normalized_content.append({**item, "url": normalized_url})
            content_changed = True

        if content_changed:
            normalized_payload["content"] = normalized_content

        return normalized_payload

    @staticmethod
    def _thread_id_from_room_message(*, message: RoomMessage) -> str | None:
        if message.type != "agent-message":
            return LegacyChatChannel._thread_id_from_room_message(message=message)

        try:
            envelope = _AgentMessageEnvelope.model_validate(message.message)
        except Exception:
            return None

        raw_thread_id = envelope.payload.get("thread_id")
        if not isinstance(raw_thread_id, str):
            return None

        thread_id = raw_thread_id.strip()
        if thread_id == "":
            return None

        return thread_id

    def _agent_message_from_room_message(
        self,
        *,
        message: RoomMessage,
    ) -> AgentMessage | None:
        if message.type != "agent-message":
            logger.debug(
                "ignoring chat room message of type %s because it has no agent equivalent",
                message.type,
            )
            return None

        envelope = _AgentMessageEnvelope.model_validate(message.message)
        payload = self._normalize_agent_message_payload(payload=envelope.payload)
        message_type = payload.get("type")
        if not isinstance(message_type, str):
            raise ValueError("agent-message payload must include a string type")

        model = _INBOUND_AGENT_MESSAGE_MODELS.get(message_type)
        if model is None:
            raise ValueError(f"unsupported agent-message payload type: {message_type}")

        return model.model_validate(payload)



from .agent import SingleRoomAgent
from meshagent.api.chan import Chan
from meshagent.api import RoomMessage, RoomClient
from meshagent.agents import AgentSessionContext
from meshagent.agents.context import TaskContext
from meshagent.tools import (
    RemoteToolkit,
    FunctionTool,
    Toolkit,
    make_toolkits,
    ToolkitBuilder,
)
from .adapter import LLMAdapter
import asyncio
import contextlib
from typing import Literal, Optional
import json
from meshagent.tools import ToolContext
import logging

from pathlib import Path
from meshagent.agents.skills import to_prompt
from meshagent.openai.tools.completions_adapter import OpenAICompletionsAdapter
from meshagent.openai import OpenAIResponsesAdapter

from .completions_thread_adapter import CompletionsThreadAdapter
from .responses_thread_adapter import ResponsesThreadAdapter
from .thread_adapter import ThreadAdapter
from .threaded_task_runner import ThreadedTaskRunner, ThreadingMode

logger = logging.getLogger("worker")
InitialMessageMode = Literal["summary", "code", "none"]


def _summarize_worker_message(*, message: object) -> str:
    if not isinstance(message, dict):
        return f"type={type(message).__name__}"

    keys = sorted(str(key) for key in message.keys())
    if len(keys) > 8:
        shown_keys = [*keys[:8], f"+{len(keys) - 8} more"]
    else:
        shown_keys = keys

    summary: list[str] = [f"keys={shown_keys}"]

    prompt_value = message.get("prompt")
    if isinstance(prompt_value, str):
        prompt_preview = " ".join(prompt_value.split())
        if len(prompt_preview) > 120:
            prompt_preview = f"{prompt_preview[:117]}..."
        summary.append(f"prompt={prompt_preview!r}")

    body_value = message.get("body")
    if isinstance(body_value, str):
        summary.append(f"body_len={len(body_value)}")

    return ", ".join(summary)


class _WorkerThreadingHelper(ThreadedTaskRunner):
    def __init__(
        self,
        *,
        threading_mode: Optional[ThreadingMode],
        thread_dir: str,
        thread_name_rules: Optional[list[str]],
        thread_name_adapter: Optional[LLMAdapter],
        thread_adapter_type: type[ThreadAdapter],
    ):
        super().__init__(
            input_schema={
                "type": "object",
                "additionalProperties": True,
                "required": [],
                "properties": {},
            },
            threading_mode=threading_mode,
            thread_dir=thread_dir,
            thread_name_rules=thread_name_rules,
            thread_name_adapter=thread_name_adapter,
            thread_adapter_type=thread_adapter_type,
        )


class SubmitWork(FunctionTool):
    def __init__(self, *, agent: "Worker", queue: str):
        self.queue = queue
        self.agent = agent
        super().__init__(
            name=f"queue_{agent.name}_task",
            title=f"Queue {agent.title} Task",
            description=f"Queues a new task to the worker -- {agent.description}",
            input_schema={
                "type": "object",
                "required": ["prompt"],
                "additionalProperties": False,
                "properties": {
                    "prompt": {
                        "type": "string",
                    },
                },
            },
        )

    async def execute(self, context: ToolContext, *, prompt: str):
        await context.room.queues.send(
            name=self.queue,
            message={
                "prompt": prompt,
            },
            create=True,
        )
        return None


class Worker(SingleRoomAgent):
    def __init__(
        self,
        *,
        queue: str,
        name=None,
        title=None,
        description=None,
        requires=None,
        llm_adapter: LLMAdapter,
        toolkits: Optional[list[Toolkit]] = None,
        rules: Optional[list[str]] = None,
        toolkit_name: Optional[str] = None,
        skill_dirs: Optional[list[str]] = None,
        supports_context: bool = True,
        annotations: Optional[list[str]] = None,
        threading_mode: Optional[ThreadingMode] = None,
        thread_dir: str = ".threads",
        thread_name_rules: Optional[list[str]] = None,
        thread_name_adapter: Optional[LLMAdapter] = None,
        initial_message_mode: InitialMessageMode = "code",
        initial_message_from: str = "worker",
        decision_model: Optional[str] = None,
        decision_llm_adapter: Optional[LLMAdapter] = None,
    ):
        super().__init__(
            name=name,
            title=title,
            description=description,
            requires=requires,
            annotations=annotations,
        )

        self._skill_dirs = skill_dirs

        self._queue = queue

        if toolkits is None:
            toolkits = []

        self._llm_adapter = llm_adapter
        self._initial_message_mode: InitialMessageMode = initial_message_mode
        normalized_initial_message_from = initial_message_from.strip()
        if normalized_initial_message_from == "":
            raise ValueError("initial_message_from must not be empty")
        self._initial_message_from = normalized_initial_message_from
        self._decision_model = (
            decision_model.strip()
            if isinstance(decision_model, str) and decision_model.strip() != ""
            else None
        )
        if self._initial_message_mode == "summary":
            self._decision_llm_adapter = (
                decision_llm_adapter
                if decision_llm_adapter is not None
                else OpenAIResponsesAdapter()
            )
        else:
            self._decision_llm_adapter = decision_llm_adapter
        resolved_thread_name_adapter = (
            thread_name_adapter if thread_name_adapter is not None else llm_adapter
        )
        thread_adapter_type: type[ThreadAdapter] = ResponsesThreadAdapter
        if isinstance(llm_adapter, OpenAICompletionsAdapter):
            thread_adapter_type = CompletionsThreadAdapter
        self._threading_helper = _WorkerThreadingHelper(
            threading_mode=threading_mode,
            thread_dir=thread_dir,
            thread_name_rules=thread_name_rules,
            thread_name_adapter=resolved_thread_name_adapter,
            thread_adapter_type=thread_adapter_type,
        )

        self._message_channel = Chan[RoomMessage]()

        self._room: RoomClient | None = None
        self._toolkits = toolkits

        if rules is None:
            rules = []

        self._rules = rules
        self._done = False

        if toolkit_name is not None:
            logger.info(f"worker will start toolkit {toolkit_name}")
            self._worker_toolkit = RemoteToolkit(
                name=toolkit_name,
                tools=[
                    SubmitWork(queue=self._queue, agent=self),
                ],
            )
        else:
            self._worker_toolkit = None

        self.supports_context = supports_context

    def _serialize_initial_message_payload(
        self,
        *,
        message: dict,
        prompt: str,
    ) -> tuple[str, str]:
        prompt_from_message = message.get("prompt")
        payload_value: object
        if isinstance(prompt_from_message, str) and prompt_from_message.strip() != "":
            payload_value = prompt_from_message
        elif prompt.strip() != "":
            payload_value = prompt
        else:
            payload_value = message

        if isinstance(payload_value, str):
            normalized_payload = payload_value.strip()
            if normalized_payload != "":
                try:
                    parsed_payload = json.loads(normalized_payload)
                except json.JSONDecodeError:
                    return payload_value, "text"

                return (
                    json.dumps(parsed_payload, indent=2, ensure_ascii=False),
                    "json",
                )

            return payload_value, "text"

        try:
            return (
                json.dumps(payload_value, indent=2, ensure_ascii=False, default=str),
                "json",
            )
        except TypeError:
            return str(payload_value), "text"

    def _format_initial_message_code_block(
        self,
        *,
        payload_text: str,
        language: str,
    ) -> str:
        normalized_language = language.strip().lower()
        if normalized_language == "":
            normalized_language = "text"
        return f"```{normalized_language}\n{payload_text.rstrip()}\n```"

    async def _summarize_initial_message_payload(
        self,
        *,
        payload_text: str,
        language: str,
    ) -> Optional[str]:
        if self._decision_llm_adapter is None:
            return None

        decision_context = self._decision_llm_adapter.create_session()
        async with decision_context:
            decision_context.append_rules(
                [
                    "Summarize worker queue requests, in a manner suitable for a conversation.",
                    "Keep it concise and factual.",
                    "Return only key details useful for later debugging.",
                    "Do not include markdown code fences in the summary.",
                ]
            )
            decision_context.append_user_message(
                (f"Summarize what this request is:\n```{language}\n{payload_text}\n```")
            )
            response = await self._decision_llm_adapter.next(
                context=decision_context,
                room=self.room,
                toolkits=[],
                model=self._decision_model
                or self._decision_llm_adapter.default_model(),
                output_schema={
                    "type": "object",
                    "required": ["summary"],
                    "additionalProperties": False,
                    "properties": {
                        "summary": {
                            "type": "string",
                            "description": (
                                "concise summary, suitable for a conversation"
                            ),
                        },
                    },
                },
            )

        if not isinstance(response, dict):
            return None

        summary = response.get("summary")
        if not isinstance(summary, str):
            return None
        normalized_summary = summary.strip()
        if normalized_summary == "":
            return None
        return normalized_summary

    async def _build_initial_thread_message(
        self,
        *,
        message: dict,
        prompt: str,
    ) -> Optional[str]:
        if self._initial_message_mode == "none":
            return None

        payload_text, payload_language = self._serialize_initial_message_payload(
            message=message,
            prompt=prompt,
        )
        if payload_text.strip() == "":
            return None

        if self._initial_message_mode == "summary":
            try:
                summary = await self._summarize_initial_message_payload(
                    payload_text=payload_text,
                    language=payload_language,
                )
            except Exception as ex:
                logger.warning(
                    "unable to summarize initial worker payload", exc_info=ex
                )
                summary = None
            if summary is not None:
                return summary

        return self._format_initial_message_code_block(
            payload_text=payload_text,
            language=payload_language,
        )

    async def preflight_start(self, *, room: RoomClient) -> None:
        del room

    async def start(self, *, room: RoomClient):
        self._done = False

        worker_toolkit_started = False
        room_agent_started = False
        try:
            if self._worker_toolkit is not None:
                await self._worker_toolkit.start(room=room)
                worker_toolkit_started = True

            await super().start(room=room)
            room_agent_started = True

            await self.preflight_start(room=room)

            self._main_task = asyncio.create_task(self.run(room=room))
        except Exception:
            self._done = True

            if room_agent_started:
                with contextlib.suppress(Exception):
                    await super().stop()

            if worker_toolkit_started and self._worker_toolkit is not None:
                with contextlib.suppress(Exception):
                    await self._worker_toolkit.stop()
            raise

    async def stop(self):
        self._done = True

        await asyncio.gather(self._main_task)

        if self._worker_toolkit is not None:
            await self._worker_toolkit.stop()

        await super().stop()

    async def get_rules(self):
        rules = [*self._rules]

        if self._skill_dirs is not None and len(self._skill_dirs) > 0:
            rules.append(
                "You have access to to following skills which follow the agentskills spec:"
            )
            rules.append(await to_prompt([*(Path(p) for p in self._skill_dirs)]))
            rules.append(
                "Use the shell or storage tool to find out more about skills and execute them when they are required"
            )

        return rules

    def get_prompt_for_message(self, *, message: dict) -> str:
        prompt = message.get("prompt")
        if prompt is None:
            logger.warning(
                "prompt property not found on worker message, inserting whole message into context"
            )
            prompt = json.dumps(message)

        return prompt

    async def append_message_context(
        self, *, message: dict, chat_context: AgentSessionContext
    ):
        if self.supports_context:
            caller_context_json = message.get("caller_context")
            if caller_context_json is not None:
                caller_context = AgentSessionContext.from_json(caller_context_json)

                chat_context.messages.extend(caller_context.messages)
                chat_context.previous_response_id = caller_context.previous_response_id

        prompt = self.get_prompt_for_message(message=message)

        chat_context.append_user_message(message=prompt)

    async def process_message(
        self,
        *,
        chat_context: AgentSessionContext,
        message: dict,
        toolkits: list[Toolkit],
    ):
        prompt = self.get_prompt_for_message(message=message)
        model = message.get("model", self._llm_adapter.default_model())
        if not isinstance(model, str) or model.strip() == "":
            model = self._llm_adapter.default_model()

        task_context = TaskContext(
            session=chat_context,
            room=self.room,
            caller=None,
            on_behalf_of=None,
            toolkits=[],
        )

        adapter_arguments = message
        thread_path = await self._threading_helper.resolve_thread_path(
            context=task_context,
            arguments=message,
            prompt=prompt,
            model=model,
        )
        if thread_path is not None:
            adapter_arguments = {
                **message,
                "path": thread_path,
            }
            await self._threading_helper.record_thread_in_index(
                context=task_context,
                path=thread_path,
            )

        thread_adapter = self._threading_helper.create_thread_adapter(
            context=task_context,
            arguments=adapter_arguments,
            attachment=None,
        )

        if thread_adapter is not None:
            await thread_adapter.start()
            self._threading_helper.ensure_local_member_on_thread(
                context=task_context,
                thread_adapter=thread_adapter,
            )
            thread_adapter.append_messages(context=chat_context)
            initial_message = await self._build_initial_thread_message(
                message=message,
                prompt=prompt,
            )
            if initial_message is not None:
                thread_adapter.write_text_message(
                    text=initial_message,
                    participant=self._initial_message_from,
                )

        def push(event: dict) -> None:
            if thread_adapter is not None:
                thread_adapter.push(event=event)

        try:
            await self.append_message_context(
                message=message, chat_context=chat_context
            )

            return await self._llm_adapter.next(
                context=chat_context,
                room=self.room,
                toolkits=toolkits,
                event_handler=push if thread_adapter is not None else None,
                model=model,
            )
        finally:
            if thread_adapter is not None:
                with contextlib.suppress(Exception):
                    await thread_adapter.stop()

    def get_toolkit_builders(self) -> list[ToolkitBuilder]:
        return []

    async def get_message_toolkits(self, *, message: dict) -> list[Toolkit]:
        toolkits = await self.get_required_toolkits(
            context=ToolContext(
                room=self.room,
                caller=self.room.local_participant,
                on_behalf_of=None,
            )
        )

        tool_providers = [*self.get_toolkit_builders()]

        model = message.get("model", self._llm_adapter.default_model())

        message_tools = message.get("tools")

        if message_tools is not None and len(message_tools) > 0:
            toolkits.extend(
                await make_toolkits(
                    room=self.room,
                    model=model,
                    providers=tool_providers,
                    tools=message_tools,
                )
            )
        return [*self._toolkits, *toolkits]

    def prepare_chat_context(self, *, chat_context: AgentSessionContext):
        pass

    async def init_session(self) -> AgentSessionContext:
        context = self._llm_adapter.create_session()
        context.append_rules(self._rules)
        return context

    async def run(self, *, room: RoomClient):
        backoff = 0
        while not self._done:
            try:
                message = await room.queues.receive(
                    name=self._queue, create=True, wait=True
                )

                backoff = 0
                if message is not None:
                    logger.info("received message on worker queue")
                    try:
                        session_context = await self.init_session()
                        async with session_context:
                            session_context.replace_rules(
                                rules=[
                                    *await self.get_rules(),
                                ]
                            )

                            toolkits = await self.get_message_toolkits(message=message)

                            self.prepare_chat_context(chat_context=session_context)

                            await self.process_message(
                                chat_context=session_context,
                                message=message,
                                toolkits=toolkits,
                            )

                    except Exception as e:
                        logger.error(
                            "Failed to process worker message: %s (%s)",
                            e,
                            _summarize_worker_message(message=message),
                            exc_info=e,
                        )

            except Exception as e:
                logger.error(
                    f"Worker error while receiving: {e}, will retry", exc_info=e
                )

                await asyncio.sleep(0.1 * pow(2, backoff))
                backoff = backoff + 1



import pytest

from meshagent.api import RoomException
from typing import Optional
from meshagent.agents.adapter import LLMAdapter
from meshagent.agents.context import AgentSessionContext
from meshagent.agents.image_captioner import LLMImageCaptioner


class _ImageCapableContext(AgentSessionContext):
    @property
    def supports_images(self) -> bool:
        return True

    @property
    def supports_files(self) -> bool:
        return False

    def append_image_message(self, *, mime_type: str, data: bytes) -> dict:
        message = {
            "role": "user",
            "content": [
                {
                    "type": "input_image",
                    "mime_type": mime_type,
                    "size": len(data),
                }
            ],
        }
        self.messages.append(message)
        return message


class _FakeAdapter(LLMAdapter):
    def __init__(self, *, response: object, image_supported: bool = True):
        self._response = response
        self._image_supported = image_supported
        self.last_context: AgentSessionContext | None = None

    def default_model(self) -> str:
        return "test-model"

    def create_session(self) -> AgentSessionContext:
        if self._image_supported:
            return _ImageCapableContext(system_role=None)
        return AgentSessionContext(system_role=None)

    async def next(
        self,
        *,
        context,
        room,
        toolkits,
        output_schema=None,
        event_handler=None,
        steering_callback=None,
        model=None,
        on_behalf_of=None,
        options: Optional[dict] = None,
    ):
        del room
        del toolkits
        del output_schema
        del event_handler
        del steering_callback
        del model
        del on_behalf_of
        self.last_context = context
        return self._response


class _DummyRoom:
    pass


@pytest.mark.asyncio
async def test_llm_image_captioner_uses_context_and_extracts_caption_key() -> None:
    adapter = _FakeAdapter(response={"caption": "A cat sleeping on a desk."})
    captioner = LLMImageCaptioner(
        llm_adapter=adapter,
        rules=["describe only the visual contents"],
        prompt="Caption this image in one sentence.",
    )

    caption = await captioner.caption(
        room=_DummyRoom(),
        image_data=b"png-bytes",
        mime_type="image/png",
    )

    assert caption == "A cat sleeping on a desk."
    assert adapter.last_context is not None
    assert adapter.last_context.messages[0]["role"] == "user"
    assert (
        adapter.last_context.messages[0]["content"]
        == "Caption this image in one sentence."
    )
    assert adapter.last_context.messages[1]["content"][0]["type"] == "input_image"


@pytest.mark.asyncio
async def test_llm_image_captioner_requires_image_capable_context() -> None:
    adapter = _FakeAdapter(response="unused", image_supported=False)
    captioner = LLMImageCaptioner(llm_adapter=adapter)

    with pytest.raises(RoomException):
        await captioner.caption(
            room=_DummyRoom(),
            image_data=b"img",
            mime_type="image/png",
        )



import logging
import asyncio
from typing import Optional

from meshagent.agents import TaskRunner
from meshagent.api.schema_document import Element, Text
from meshagent.api.room_server_client import RoomClient, MeshDocument
from meshagent.agents.task_runner import TaskContext

logger = logging.getLogger(__name__)


class ListenerContext:
    def __init__(
        self, document: MeshDocument, room: RoomClient, call_context: TaskContext
    ):
        self.document = document
        self.call_context = call_context
        self.room = room


# Notifies of new nodes or changed nodes in a document, the document must already exist
class Listener(TaskRunner):
    def __init__(
        self,
        *,
        name: str,
        wait_for_synchronize: bool = True,
        title: Optional[str] = None,
        description: Optional[str] = None,
    ):
        super().__init__(
            name=name,
            description=description,
            title=title,
            input_schema={
                "type": "object",
                "required": ["path"],
                "additionalProperties": False,
                "properties": {
                    "path": {
                        "type": "string",
                        "description": "the path of the document to listen to",
                    }
                },
            },
            output_schema={
                "type": "object",
                "additionalProperties": False,
                "required": [],
                "properties": {},
            },
        )
        self.wait_for_synchronize = wait_for_synchronize

    async def on_listening_started(self, listener_context: ListenerContext):
        pass

    async def on_element_inserted(
        self, listener_context: ListenerContext, element: Element
    ) -> bool:
        return False

    async def on_attribute_changed(
        self,
        listener_context: ListenerContext,
        element: Element,
        attribute: Optional[str],
    ) -> bool:
        return False

    async def ask(self, *, context: TaskContext, arguments: dict):
        output_path = arguments["path"]
        room = context.room
        logger.info("Visitor connecting to %s", output_path)
        doc = await room.sync.open(path=output_path, create=True)
        try:
            listener_context = ListenerContext(
                document=doc,
                room=room,
                call_context=context,
            )
            logger.info("Visitor connected to %s", output_path)

            change_queue = list[Element | Text]()

            def append_children(node: Element):
                for child in node.get_children():
                    if child not in change_queue:
                        change_queue.append([child])
                    if isinstance(child, Element):
                        append_children(child)

            if not self.wait_for_synchronize:
                change_queue.append([doc.root])
                append_children(doc.root)
            else:
                await doc.synchronized

            await self.on_listening_started(listener_context=listener_context)

            wait_for_changes = asyncio.Future()

            @doc.on("inserted")
            def on_inserted(e: Element):
                logger.info("element inserted %s", e.tag_name)
                if e not in change_queue:
                    change_queue.append([e])
                    append_children(e)

                if not wait_for_changes.done():
                    wait_for_changes.set_result(True)

            @doc.on("updated")
            def on_updated(e: Element, attribute: str):
                logger.info("element updated %s", e.tag_name)
                if e not in change_queue:
                    change_queue.append([e, attribute])
                    #
                    # append_children(e)

                if not wait_for_changes.done():
                    wait_for_changes.set_result(True)

            waiting_for_end = True
            while waiting_for_end:
                await wait_for_changes

                while len(change_queue) > 0:
                    change = change_queue.pop(0)
                    content = change[0]
                    if len(change) > 1:
                        done = await self.on_attribute_changed(
                            listener_context, content, change[1]
                        )
                        if done:
                            waiting_for_end = False

                    else:
                        done = await self.on_element_inserted(listener_context, content)
                        if done:
                            waiting_for_end = False

                    if content in change_queue:
                        change_queue.remove(content)

                wait_for_changes = asyncio.Future()
        except Exception as e:
            logger.error("Failed to visit", exc_info=e)
            raise

        finally:
            logger.info("vistor done")

            await asyncio.sleep(5)
            await room.sync.close(output_path)

            return {}



from __future__ import annotations

import logging
import re
import uuid
from pathlib import PurePosixPath
from typing import Any, TypeVar
from urllib.parse import urlparse

from meshagent.api import (
    Element,
    Participant,
    RoomClient,
    RoomException,
    RoomMessage,
)
from meshagent.api.messaging import JsonContent
from pydantic import BaseModel, ValidationError
from meshagent.tools import (
    FunctionTool,
    RemoteToolkit,
    ToolContext,
    Toolkit,
    ToolkitBuilder,
    tool,
)
from meshagent.tools.strict_schema import ensure_strict_json_schema

from .adapter import LLMAdapter
from .messages import (
    AGENT_EVENT_THREAD_CLEARED,
    AGENT_EVENT_FILE_CONTENT_DELTA,
    AGENT_EVENT_FILE_CONTENT_ENDED,
    AGENT_EVENT_FILE_CONTENT_STARTED,
    AGENT_EVENT_TOOL_CALL_APPROVAL_REQUESTED,
    AGENT_EVENT_TEXT_CONTENT_DELTA,
    AGENT_EVENT_TEXT_CONTENT_ENDED,
    AGENT_EVENT_TEXT_CONTENT_STARTED,
    AGENT_EVENT_TURN_ENDED,
    AGENT_EVENT_TURN_STARTED,
    AGENT_MESSAGE_THREAD_CLEAR,
    AGENT_MESSAGE_TOOL_CALL_APPROVE,
    AGENT_MESSAGE_TOOL_CALL_REJECT,
    AGENT_MESSAGE_TURN_INTERRUPT,
    AGENT_MESSAGE_TURN_START,
    AGENT_MESSAGE_TURN_STEER,
    AgentFileContent,
    AgentFileContentDelta,
    AgentFileContentEnded,
    AgentFileContentStarted,
    AgentMessage,
    AgentTextContent,
    AgentTextContentDelta,
    AgentTextContentEnded,
    AgentTextContentStarted,
    AgentToolCallApprovalRequested,
    ApproveAgentToolCall,
    ClearThread,
    RejectAgentToolCall,
    ThreadCleared,
    TurnEnded,
    TurnInterrupt,
    TurnStart,
    TurnStarted,
    TurnSteer,
)
from .process import Message
from .threaded_channel import ThreadedChannel
from .toolkit_schema import build_tools_property_schema

logger = logging.getLogger("legacy-chat-channel")
_MessageT = TypeVar("_MessageT", bound=AgentMessage)


class _ChatAttachmentPayload(BaseModel):
    path: str


class _ChatMessagePayload(BaseModel):
    path: str
    text: str = ""
    attachments: list[_ChatAttachmentPayload] | None = None
    tools: list[dict[str, Any]] | None = None
    model: str | None = None


class _PathMessagePayload(BaseModel):
    path: str


class _ApprovalDecisionPayload(BaseModel):
    path: str
    approval_id: str


class LegacyChatChannel(ThreadedChannel):
    def __init__(
        self,
        *,
        room: RoomClient,
        threading_mode: str | None = None,
        thread_dir: str | None = None,
        llm_adapter: LLMAdapter | None = None,
        empty_state_title: str = "How can I help you?",
        toolkit_builders: list[ToolkitBuilder] | None = None,
    ) -> None:
        super().__init__(
            room=room,
            threading_mode=threading_mode,
            thread_dir=thread_dir,
            llm_adapter=llm_adapter,
        )
        self._empty_state_title = empty_state_title
        self._toolkit_builders = list(toolkit_builders or [])
        self._active_turn_ids_by_thread: dict[str, str] = {}
        self._pending_approval_turn_ids_by_thread: dict[str, dict[str, str]] = {}
        self._open_participant_ids_by_thread: dict[str, set[str]] = {}
        self._active_text_by_thread: dict[str, dict[str, str]] = {}
        self._active_files_by_thread: dict[str, dict[str, str]] = {}

    def _uses_explicit_thread_dir_for_thread_list(self) -> bool:
        return True

    def handles(self, message: Message) -> bool:
        return message.data.type in {
            AGENT_EVENT_FILE_CONTENT_STARTED,
            AGENT_EVENT_FILE_CONTENT_DELTA,
            AGENT_EVENT_FILE_CONTENT_ENDED,
            AGENT_EVENT_TEXT_CONTENT_STARTED,
            AGENT_EVENT_TEXT_CONTENT_DELTA,
            AGENT_EVENT_TEXT_CONTENT_ENDED,
            AGENT_EVENT_THREAD_CLEARED,
            AGENT_EVENT_TURN_STARTED,
            AGENT_EVENT_TURN_ENDED,
            AGENT_EVENT_TOOL_CALL_APPROVAL_REQUESTED,
        }

    async def on_start(self) -> None:
        self._room.messaging.on("message", self._on_room_message)
        await self.publish_thread_attributes()
        await self._room.local_participant.set_attribute(
            "empty_state_title",
            self._empty_state_title,
        )
        await self.open_thread_list_document()
        if not self._room.messaging.is_enabled:
            await self._room.messaging.enable()

    async def on_stop(self) -> None:
        self._room.messaging.off("message", self._on_room_message)
        await self.close_thread_list_document()
        self._active_turn_ids_by_thread.clear()
        self._pending_approval_turn_ids_by_thread.clear()
        self._open_participant_ids_by_thread.clear()
        self._active_text_by_thread.clear()
        self._active_files_by_thread.clear()

    async def on_message(self, message: Message) -> None:
        data = message.data
        if data.type == AGENT_EVENT_TEXT_CONTENT_STARTED:
            text_started = self._coerce_message(
                data=data, model=AgentTextContentStarted
            )
            active_text = self._active_text_by_thread.setdefault(
                text_started.thread_id,
                {},
            )
            active_text[text_started.item_id] = ""
            return

        if data.type == AGENT_EVENT_TEXT_CONTENT_DELTA:
            text_delta = self._coerce_message(data=data, model=AgentTextContentDelta)
            active_text = self._active_text_by_thread.setdefault(
                text_delta.thread_id,
                {},
            )
            current_text = active_text.get(text_delta.item_id, "")
            active_text[text_delta.item_id] = current_text + text_delta.text
            return

        if data.type == AGENT_EVENT_TEXT_CONTENT_ENDED:
            text_ended = self._coerce_message(data=data, model=AgentTextContentEnded)
            active_text = self._active_text_by_thread.get(text_ended.thread_id, {})
            text = active_text.pop(text_ended.item_id, "")
            if len(active_text) == 0:
                self._active_text_by_thread.pop(text_ended.thread_id, None)
            if text.strip() != "":
                self._send_chat_to_open_participants(
                    thread_id=text_ended.thread_id,
                    text=text,
                )
            return

        if data.type == AGENT_EVENT_FILE_CONTENT_STARTED:
            file_started = self._coerce_message(
                data=data, model=AgentFileContentStarted
            )
            active_files = self._active_files_by_thread.setdefault(
                file_started.thread_id,
                {},
            )
            active_files[file_started.item_id] = ""
            return

        if data.type == AGENT_EVENT_FILE_CONTENT_DELTA:
            file_delta = self._coerce_message(data=data, model=AgentFileContentDelta)
            active_files = self._active_files_by_thread.setdefault(
                file_delta.thread_id,
                {},
            )
            active_files[file_delta.item_id] = file_delta.url
            return

        if data.type == AGENT_EVENT_FILE_CONTENT_ENDED:
            file_ended = self._coerce_message(data=data, model=AgentFileContentEnded)
            active_files = self._active_files_by_thread.get(file_ended.thread_id, {})
            path = active_files.pop(file_ended.item_id, "")
            if len(active_files) == 0:
                self._active_files_by_thread.pop(file_ended.thread_id, None)
            if path.strip() != "":
                self._send_attachments_to_open_participants(
                    thread_id=file_ended.thread_id,
                    attachments=[path],
                )
            return

        if data.type == AGENT_EVENT_TURN_STARTED:
            turn_started = self._coerce_message(data=data, model=TurnStarted)
            self._active_turn_ids_by_thread[turn_started.thread_id] = (
                turn_started.turn_id
            )
            return

        if data.type == AGENT_EVENT_THREAD_CLEARED:
            thread_cleared = self._coerce_message(data=data, model=ThreadCleared)
            self._clear_tracked_thread_state(thread_id=thread_cleared.thread_id)
            self._send_cleared_to_open_participants(thread_id=thread_cleared.thread_id)
            return

        if data.type == AGENT_EVENT_TURN_ENDED:
            turn_ended = self._coerce_message(data=data, model=TurnEnded)
            tracked_turn_id = self._active_turn_ids_by_thread.get(turn_ended.thread_id)
            if tracked_turn_id == turn_ended.turn_id:
                self._active_turn_ids_by_thread.pop(turn_ended.thread_id, None)
            self._active_text_by_thread.pop(turn_ended.thread_id, None)
            self._active_files_by_thread.pop(turn_ended.thread_id, None)

            pending_approvals = self._pending_approval_turn_ids_by_thread.get(
                turn_ended.thread_id
            )
            if pending_approvals is None:
                return

            remaining_approvals = {
                approval_id: pending_turn_id
                for approval_id, pending_turn_id in pending_approvals.items()
                if pending_turn_id != turn_ended.turn_id
            }
            if len(remaining_approvals) == 0:
                self._pending_approval_turn_ids_by_thread.pop(
                    turn_ended.thread_id,
                    None,
                )
            else:
                self._pending_approval_turn_ids_by_thread[turn_ended.thread_id] = (
                    remaining_approvals
                )
            return

        if data.type == AGENT_EVENT_TOOL_CALL_APPROVAL_REQUESTED:
            approval = self._coerce_message(
                data=data,
                model=AgentToolCallApprovalRequested,
            )
            pending_approvals = self._pending_approval_turn_ids_by_thread.setdefault(
                approval.thread_id,
                {},
            )
            pending_approvals[approval.item_id] = approval.turn_id

    def _on_room_message(self, *, message: RoomMessage) -> None:
        sender = self._room.messaging.get_participant(message.from_participant_id)
        if sender is None:
            logger.warning(
                "ignoring chat message from unknown participant %s",
                message.from_participant_id,
            )
            return

        if self._handle_room_control_message(message=message, sender=sender):
            return

        thread_id = self._thread_id_from_room_message(message=message)
        if thread_id is not None:
            self._register_open_participant(
                thread_id=thread_id,
                participant_id=sender.id,
            )
            if self._should_touch_thread_index_for_room_message(
                message_type=message.type
            ):
                self.bump_thread(path=thread_id)

        try:
            agent_message = self._agent_message_from_room_message(message=message)
        except ValidationError:
            logger.exception(
                "unable to translate chat room message of type %s",
                message.type,
            )
            return

        if agent_message is None:
            return

        self.emit(sender=sender, payload=agent_message)

    @staticmethod
    def _coerce_message(
        *,
        data: AgentMessage,
        model: type[_MessageT],
    ) -> _MessageT:
        if isinstance(data, model):
            return data
        return model.model_validate(data.model_dump(mode="python"))

    @staticmethod
    def _normalize_attachment_url(*, path: str) -> str | None:
        normalized_path = path.strip()
        if normalized_path == "":
            return None

        parsed = urlparse(normalized_path)
        if parsed.scheme != "":
            return normalized_path

        room_path = normalized_path.lstrip("/")
        if room_path == "":
            return None

        return f"room:///{room_path}"

    @staticmethod
    def _room_storage_path_from_attachment_url(*, url: str) -> str | None:
        parsed = urlparse(url)
        if parsed.scheme == "":
            raw_path = url.strip()
        elif parsed.scheme == "room":
            raw_path = f"{parsed.netloc}{parsed.path}"
        else:
            return None

        normalized = PurePosixPath("/" + raw_path).as_posix().strip("/")
        if normalized == "":
            return None

        if any(part in {".", ".."} for part in PurePosixPath(normalized).parts):
            return None

        return normalized

    @staticmethod
    def _content_from_chat_message(
        *,
        payload: _ChatMessagePayload,
    ) -> list[AgentTextContent | AgentFileContent]:
        content: list[AgentTextContent | AgentFileContent] = []

        if payload.text.strip() != "":
            content.append(
                AgentTextContent(
                    type="text",
                    text=payload.text,
                )
            )

        if payload.attachments is not None:
            for attachment in payload.attachments:
                normalized_url = LegacyChatChannel._normalize_attachment_url(
                    path=attachment.path
                )
                if normalized_url is None:
                    continue
                content.append(
                    AgentFileContent(
                        type="file",
                        url=normalized_url,
                    )
                )

        return content

    def _active_turn_id(self, *, thread_id: str) -> str | None:
        return self._active_turn_ids_by_thread.get(thread_id)

    def _thread_and_turn_id_from_tool_context(
        self, *, context: ToolContext
    ) -> tuple[str, str]:
        caller_context = context.caller_context
        if not isinstance(caller_context, dict):
            raise RoomException(
                "chat tool requires thread_id and turn_id in caller_context"
            )

        raw_thread_id = caller_context.get("thread_id")
        if not isinstance(raw_thread_id, str) or raw_thread_id.strip() == "":
            raise RoomException("chat tool requires a non-empty thread_id")
        thread_id = raw_thread_id.strip()

        raw_turn_id = caller_context.get("turn_id")
        if isinstance(raw_turn_id, str) and raw_turn_id.strip() != "":
            return thread_id, raw_turn_id.strip()

        turn_id = self._active_turn_id(thread_id=thread_id)
        if turn_id is None:
            raise RoomException("attach_file requires an active turn")

        return thread_id, turn_id

    @staticmethod
    def _thread_id_from_room_message(*, message: RoomMessage) -> str | None:
        payload = message.message
        if not isinstance(payload, dict):
            return None

        path = payload.get("path")
        if not isinstance(path, str):
            return None

        normalized_path = path.strip()
        if normalized_path == "":
            return None

        return normalized_path

    def _build_thread_list_tools(self) -> list[FunctionTool]:
        if self._thread_list_dir() is None:
            return []

        read_file_hint = (
            "Use read_file with a thread path to read that thread's contents."
        )
        outer = self

        def to_json_entry(entry: Element) -> dict[str, str]:
            return {
                "name": str(entry.get_attribute("name") or ""),
                "path": str(entry.get_attribute("path") or ""),
                "modified_at": str(entry.get_attribute("modified_at") or ""),
                "created_at": str(entry.get_attribute("created_at") or ""),
            }

        @tool(
            name="list_threads",
            description="lists recent threads sorted by last modified date (newest first). Use read_file with a thread path to read that thread's contents.",
        )
        def list_threads(*, limit: int = 20, offset: int = 0) -> JsonContent:
            normalized_offset = max(0, int(offset))
            normalized_limit = max(1, min(200, int(limit)))

            entries = outer._sorted_thread_list_entries()
            if len(entries) == 0:
                return JsonContent(
                    json={
                        "threads": [],
                        "total": 0,
                        "offset": normalized_offset,
                        "limit": normalized_limit,
                        "message": "no threads were found in the thread list",
                        "read_file_hint": read_file_hint,
                    }
                )

            selected = outer._thread_list_slice(
                entries=entries,
                limit=limit,
                offset=offset,
            )
            if len(selected) == 0:
                return JsonContent(
                    json={
                        "threads": [],
                        "total": len(entries),
                        "offset": normalized_offset,
                        "limit": normalized_limit,
                        "message": "no threads were found for the requested limit/offset",
                        "read_file_hint": read_file_hint,
                    }
                )

            return JsonContent(
                json={
                    "threads": [to_json_entry(entry) for entry in selected],
                    "total": len(entries),
                    "offset": normalized_offset,
                    "limit": normalized_limit,
                    "sort": "modified_at_desc",
                    "read_file_hint": read_file_hint,
                }
            )

        @tool(
            name="grep_thread_list",
            description="searches the thread list for matching thread names and paths. Use read_file with a thread path to read that thread's contents.",
        )
        def grep_thread_list(*, pattern: str, ignore_case: bool = True) -> JsonContent:
            needle = pattern.strip()
            if needle == "":
                return JsonContent(
                    json={
                        "threads": [],
                        "total_matches": 0,
                        "pattern": needle,
                        "ignore_case": ignore_case,
                        "message": "pattern is required",
                        "read_file_hint": read_file_hint,
                    }
                )

            flags = re.IGNORECASE if ignore_case else 0
            try:
                matcher = re.compile(needle, flags)
            except re.error as ex:
                return JsonContent(
                    json={
                        "threads": [],
                        "total_matches": 0,
                        "pattern": needle,
                        "ignore_case": ignore_case,
                        "error": "invalid_regex_pattern",
                        "message": f"invalid regex pattern: {ex}",
                        "read_file_hint": read_file_hint,
                    }
                )

            matches: list[dict[str, str]] = []
            for entry in outer._sorted_thread_list_entries():
                name = entry.get_attribute("name")
                path = entry.get_attribute("path")
                created_at = entry.get_attribute("created_at")
                modified_at = entry.get_attribute("modified_at")
                haystack = f"{name}\n{path}\n{created_at}\n{modified_at}"
                if matcher.search(haystack) is None:
                    continue
                matches.append(to_json_entry(entry))

            if len(matches) == 0:
                return JsonContent(
                    json={
                        "threads": [],
                        "total_matches": 0,
                        "pattern": needle,
                        "ignore_case": ignore_case,
                        "message": "no matching threads were found",
                        "read_file_hint": read_file_hint,
                    }
                )

            return JsonContent(
                json={
                    "threads": matches,
                    "total_matches": len(matches),
                    "pattern": needle,
                    "ignore_case": ignore_case,
                    "read_file_hint": read_file_hint,
                }
            )

        return [list_threads, grep_thread_list]

    def _local_participant_name(self) -> str:
        local_name = self._room.local_participant.get_attribute("name")
        if not isinstance(local_name, str) or local_name.strip() == "":
            return "assistant"
        return local_name.strip()

    def _build_new_thread_tool_schema(self) -> dict[str, Any]:
        tools_schema: dict[str, Any] = {
            "type": "object",
            "additionalProperties": False,
            "required": ["message"],
            "properties": {
                "message": {
                    "type": "object",
                    "additionalProperties": False,
                    "required": ["text"],
                    "properties": {
                        "text": {"type": "string"},
                        "attachments": {
                            "anyOf": [
                                {
                                    "type": "array",
                                    "items": {"$ref": "#/$defs/ChatAttachment"},
                                },
                                {"type": "null"},
                            ]
                        },
                    },
                }
            },
            "$defs": {
                "ChatAttachment": {
                    "type": "object",
                    "additionalProperties": False,
                    "required": ["path"],
                    "properties": {
                        "path": {"type": "string"},
                    },
                }
            },
        }

        message_tools_schema, defs = build_tools_property_schema(
            toolkit_builders=self._toolkit_builders
        )
        if message_tools_schema is not None:
            tools_schema["properties"]["message"]["properties"]["tools"] = (
                message_tools_schema
            )
        else:
            tools_schema["properties"]["message"]["properties"]["tools"] = {
                "anyOf": [
                    {
                        "type": "array",
                        "items": ensure_strict_json_schema({}),
                    },
                    {"type": "null"},
                ],
            }

        for key, value in defs.items():
            tools_schema["$defs"][key] = value
        return ensure_strict_json_schema(tools_schema)

    def _make_new_thread_tool(self) -> FunctionTool:
        local_name = self._local_participant_name()
        tools_schema = self._build_new_thread_tool_schema()

        outer = self

        class NewThreadTool(FunctionTool):
            def __init__(self) -> None:
                super().__init__(
                    name="new_thread",
                    description=f"starts a new thread for {local_name}, posts a message to the thread, and then returns the path and name of the new thread. Since work will continue asynchronously on that thread, an agent should not invoke this if it will check the result of the work, it should generally be invoked as fire and forget.",
                    input_schema=tools_schema,
                    supports_context=True,
                )

            async def execute(
                self,
                context: ToolContext,
                *,
                message: dict[str, Any],
            ) -> JsonContent:
                if outer.supervisor is None:
                    raise RoomException(
                        "chat channel must be attached to a supervisor before using chat.new_thread"
                    )

                text_value = message.get("text")
                text = text_value if isinstance(text_value, str) else ""
                payload = {**message, "text": text}
                attachment_paths = [
                    attachment.path
                    for attachment in payload.get("attachments") or []
                    if isinstance(attachment, _ChatAttachmentPayload)
                ]
                if len(attachment_paths) == 0:
                    raw_attachments = payload.get("attachments")
                    if isinstance(raw_attachments, list):
                        attachment_paths = [
                            attachment_value["path"]
                            for attachment_value in raw_attachments
                            if isinstance(attachment_value, dict)
                            and isinstance(attachment_value.get("path"), str)
                            and attachment_value.get("path", "").strip() != ""
                        ]

                path, friendly_name = await outer.new_thread(
                    message_text=text,
                    attachments=attachment_paths,
                    on_behalf_of=context.on_behalf_of or context.caller,
                )

                chat_message = _ChatMessagePayload.model_validate(
                    {
                        "path": path,
                        "text": text,
                        "attachments": payload.get("attachments"),
                        "tools": payload.get("tools"),
                    }
                )
                outer.emit(
                    sender=context.on_behalf_of or context.caller,
                    payload=TurnStart(
                        type=AGENT_MESSAGE_TURN_START,
                        thread_id=path,
                        content=outer._content_from_chat_message(payload=chat_message),
                        toolkits=chat_message.tools,
                    ),
                )
                return JsonContent(json={"path": path, "name": friendly_name})

        return NewThreadTool()

    def _make_attach_file_tool(self) -> FunctionTool:
        outer = self

        @tool(
            name="attach_file",
            description="attach a room file path or URL to the current thread so the user can see it",
        )
        async def attach_file(context: ToolContext, path: str) -> None:
            thread_id, turn_id = outer._thread_and_turn_id_from_tool_context(
                context=context
            )
            normalized_url = outer._normalize_attachment_url(path=path)
            if normalized_url is None:
                raise RoomException("attach_file requires a non-empty path")

            room_storage_path = outer._room_storage_path_from_attachment_url(
                url=normalized_url
            )
            if room_storage_path is not None:
                try:
                    exists = await context.room.storage.exists(path=room_storage_path)
                except Exception as exc:
                    raise RoomException(
                        f"attach_file could not verify room file {room_storage_path}: {exc}"
                    ) from exc
                if not exists:
                    raise RoomException(
                        f"attach_file could not find a room file at {room_storage_path}"
                    )

            item_id = str(uuid.uuid4())
            sender = context.on_behalf_of or context.caller
            outer.emit(
                sender=sender,
                payload=AgentFileContentStarted(
                    type=AGENT_EVENT_FILE_CONTENT_STARTED,
                    thread_id=thread_id,
                    turn_id=turn_id,
                    item_id=item_id,
                ),
            )
            outer.emit(
                sender=sender,
                payload=AgentFileContentDelta(
                    type=AGENT_EVENT_FILE_CONTENT_DELTA,
                    thread_id=thread_id,
                    turn_id=turn_id,
                    item_id=item_id,
                    url=normalized_url,
                ),
            )
            outer.emit(
                sender=sender,
                payload=AgentFileContentEnded(
                    type=AGENT_EVENT_FILE_CONTENT_ENDED,
                    thread_id=thread_id,
                    turn_id=turn_id,
                    item_id=item_id,
                ),
            )

        return attach_file

    def _build_chat_tools(self) -> list[FunctionTool]:
        return [
            self._make_new_thread_tool(),
            self._make_attach_file_tool(),
            *self._build_thread_list_tools(),
        ]

    def get_agent_toolkits(self) -> list[Toolkit]:
        return [
            Toolkit(
                name="chat",
                tools=self._build_chat_tools(),
            )
        ]

    def make_remote_toolkit(self) -> RemoteToolkit:
        local_name = self._local_participant_name()
        return RemoteToolkit(
            name="chat",
            description=f"tools for interacting with {local_name}",
            public=False,
            tools=self._build_chat_tools(),
            validation_mode="content_types",
        )

    def _send_thread_tool_providers(
        self,
        *,
        to: Participant,
        path: str,
    ) -> None:
        self._room.messaging.send_message_nowait(
            to=to,
            type="set_thread_tool_providers",
            message={
                "path": path,
                "tool_providers": [
                    {"name": toolkit_builder.name}
                    for toolkit_builder in self._toolkit_builders
                ],
            },
        )

    def _handle_room_control_message(
        self,
        *,
        message: RoomMessage,
        sender: Participant,
    ) -> bool:
        if message.type == "get_thread_toolkit_builders":
            thread_id = self._thread_id_from_room_message(message=message)
            if thread_id is None:
                return True
            self._register_open_participant(
                thread_id=thread_id,
                participant_id=sender.id,
            )
            self._send_thread_tool_providers(
                to=sender,
                path=thread_id,
            )
            return True

        if message.type == "typing":
            return True

        return False

    @staticmethod
    def _should_touch_thread_index_for_room_message(*, message_type: str) -> bool:
        return message_type not in {
            "opened",
            "cleared",
            "get_thread_toolkit_builders",
            "typing",
        }

    def _register_open_participant(
        self,
        *,
        thread_id: str,
        participant_id: str,
    ) -> None:
        participant_ids = self._open_participant_ids_by_thread.setdefault(
            thread_id,
            set(),
        )
        participant_ids.add(participant_id)

    def _open_participants(self, *, thread_id: str) -> list[Participant]:
        participant_ids = self._open_participant_ids_by_thread.get(thread_id)
        if participant_ids is None:
            return []

        online_participants: list[Participant] = []
        stale_participant_ids: list[str] = []
        for participant_id in participant_ids:
            participant = self._room.messaging.get_participant(participant_id)
            if participant is None:
                stale_participant_ids.append(participant_id)
                continue
            online_participants.append(participant)

        for participant_id in stale_participant_ids:
            participant_ids.discard(participant_id)

        if len(participant_ids) == 0:
            self._open_participant_ids_by_thread.pop(thread_id, None)

        return online_participants

    def _approval_turn_id(self, *, thread_id: str, approval_id: str) -> str | None:
        pending_approvals = self._pending_approval_turn_ids_by_thread.get(thread_id)
        if pending_approvals is None:
            return None

        turn_id = pending_approvals.pop(approval_id, None)
        if len(pending_approvals) == 0:
            self._pending_approval_turn_ids_by_thread.pop(thread_id, None)
        return turn_id

    def _clear_tracked_thread_state(self, *, thread_id: str) -> None:
        self._active_turn_ids_by_thread.pop(thread_id, None)
        self._pending_approval_turn_ids_by_thread.pop(thread_id, None)
        self._active_text_by_thread.pop(thread_id, None)
        self._active_files_by_thread.pop(thread_id, None)

    def _send_chat_to_open_participants(self, *, thread_id: str, text: str) -> None:
        for participant in self._open_participants(thread_id=thread_id):
            if participant.id == self._room.local_participant.id:
                continue

            self._room.messaging.send_message_nowait(
                to=participant,
                type="chat",
                message={
                    "path": thread_id,
                    "text": text,
                },
            )

    def _send_attachments_to_open_participants(
        self,
        *,
        thread_id: str,
        attachments: list[str],
    ) -> None:
        normalized_attachments = [
            {"path": attachment}
            for attachment in attachments
            if isinstance(attachment, str) and attachment.strip() != ""
        ]
        if len(normalized_attachments) == 0:
            return

        for participant in self._open_participants(thread_id=thread_id):
            if participant.id == self._room.local_participant.id:
                continue

            self._room.messaging.send_message_nowait(
                to=participant,
                type="chat",
                message={
                    "path": thread_id,
                    "attachments": normalized_attachments,
                },
            )

    def _send_cleared_to_open_participants(self, *, thread_id: str) -> None:
        for participant in self._open_participants(thread_id=thread_id):
            if participant.id == self._room.local_participant.id:
                continue

            self._room.messaging.send_message_nowait(
                to=participant,
                type="cleared",
                message={
                    "path": thread_id,
                },
            )

    def _agent_message_from_room_message(
        self,
        *,
        message: RoomMessage,
    ) -> AgentMessage | None:
        message_type = message.type

        if message_type == "chat":
            payload = _ChatMessagePayload.model_validate(message.message)
            return TurnStart(
                type=AGENT_MESSAGE_TURN_START,
                thread_id=payload.path,
                content=self._content_from_chat_message(payload=payload),
                toolkits=payload.tools,
                model=payload.model,
            )

        if message_type == "steer":
            payload = _ChatMessagePayload.model_validate(message.message)
            turn_id = self._active_turn_id(thread_id=payload.path)
            if turn_id is None:
                logger.debug(
                    "ignoring steer for thread %s without an active turn",
                    payload.path,
                )
                return None

            return TurnSteer(
                type=AGENT_MESSAGE_TURN_STEER,
                thread_id=payload.path,
                turn_id=turn_id,
                content=self._content_from_chat_message(payload=payload),
                toolkits=payload.tools,
            )

        if message_type == "cancel":
            payload = _PathMessagePayload.model_validate(message.message)
            turn_id = self._active_turn_id(thread_id=payload.path)
            if turn_id is None:
                logger.debug(
                    "ignoring cancel for thread %s without an active turn",
                    payload.path,
                )
                return None

            return TurnInterrupt(
                type=AGENT_MESSAGE_TURN_INTERRUPT,
                thread_id=payload.path,
                turn_id=turn_id,
            )

        if message_type in {"approved", "rejected"}:
            payload = _ApprovalDecisionPayload.model_validate(message.message)
            turn_id = self._approval_turn_id(
                thread_id=payload.path,
                approval_id=payload.approval_id,
            )
            if turn_id is None:
                logger.debug(
                    "ignoring %s for thread %s without a pending approval",
                    message_type,
                    payload.path,
                )
                return None

            if message_type == "approved":
                return ApproveAgentToolCall(
                    type=AGENT_MESSAGE_TOOL_CALL_APPROVE,
                    thread_id=payload.path,
                    turn_id=turn_id,
                    item_id=payload.approval_id,
                )

            return RejectAgentToolCall(
                type=AGENT_MESSAGE_TOOL_CALL_REJECT,
                thread_id=payload.path,
                turn_id=turn_id,
                item_id=payload.approval_id,
            )

        if message_type == "clear":
            payload = _PathMessagePayload.model_validate(message.message)
            return ClearThread(
                type=AGENT_MESSAGE_THREAD_CLEAR,
                thread_id=payload.path,
            )

        if message_type in {"opened", "cleared"}:
            logger.debug(
                "ignoring chat room message of type %s because it has no agent equivalent",
                message_type,
            )
            return None

        return None



from typing import Any

import pytest

from meshagent.agents.agent import RoomTool
from meshagent.api import RoomException
from meshagent.api.messaging import TextContent
from meshagent.tools import ToolContext


class _FakeParticipant:
    def __init__(self, *, participant_id: str, name: str):
        self.id = participant_id
        self._name = name

    def get_attribute(self, key: str) -> Any:
        if key == "name":
            return self._name
        return None


class _FakeIterableResult:
    def __aiter__(self):
        return self

    async def __anext__(self):
        raise StopAsyncIteration


class _FakeAgentsClient:
    def __init__(self, *, response: Any):
        self._response = response
        self.calls: list[dict[str, Any]] = []

    async def invoke_tool(
        self,
        *,
        toolkit: str,
        tool: str,
        input: Any,
        participant_id: str | None = None,
        on_behalf_of_id: str | None = None,
        caller_context: dict | None = None,
    ) -> Any:
        self.calls.append(
            {
                "toolkit": toolkit,
                "tool": tool,
                "input": input,
                "participant_id": participant_id,
                "on_behalf_of_id": on_behalf_of_id,
                "caller_context": caller_context,
            }
        )
        return self._response


class _FakeRoom:
    def __init__(self, *, agents: _FakeAgentsClient):
        self.agents = agents


@pytest.mark.asyncio
async def test_room_tool_raises_if_remote_tool_returns_iterable() -> None:
    fake_agents = _FakeAgentsClient(response=_FakeIterableResult())
    room = _FakeRoom(agents=fake_agents)
    caller = _FakeParticipant(participant_id="caller-id", name="caller")
    context = ToolContext(
        room=room,
        caller=caller,
        caller_context={"chat": {"id": "chat-1"}},
    )
    tool = RoomTool(
        toolkit_name="remote_tools",
        name="computer_call",
        input_schema={
            "type": "object",
            "required": [],
            "properties": {},
            "additionalProperties": False,
        },
    )

    with pytest.raises(RoomException, match="returned an iterable stream"):
        await tool.execute(context=context)

    assert len(fake_agents.calls) == 1
    assert fake_agents.calls[0]["caller_context"] == {"chat": {"id": "chat-1"}}


@pytest.mark.asyncio
async def test_room_tool_uses_non_stream_call_without_event_handler() -> None:
    response = TextContent(text="ok")
    fake_agents = _FakeAgentsClient(response=response)
    room = _FakeRoom(agents=fake_agents)
    caller = _FakeParticipant(participant_id="caller-id", name="caller")
    context = ToolContext(room=room, caller=caller)
    tool = RoomTool(
        toolkit_name="remote_tools",
        name="computer_call",
        input_schema={
            "type": "object",
            "required": [],
            "properties": {},
            "additionalProperties": False,
        },
    )

    result = await tool.execute(context=context)

    assert result == response
    assert len(fake_agents.calls) == 1


def test_room_tool_defaults_to_strict_when_metadata_is_missing() -> None:
    tool = RoomTool(
        toolkit_name="remote_tools",
        name="computer_call",
        input_schema={
            "type": "object",
            "required": [],
            "properties": {},
            "additionalProperties": False,
        },
    )

    assert tool.strict is True


def test_room_tool_preserves_explicit_non_strict_metadata() -> None:
    tool = RoomTool(
        toolkit_name="remote_tools",
        name="computer_call",
        input_schema={
            "type": "object",
            "required": [],
            "properties": {},
            "additionalProperties": False,
        },
        strict=False,
    )

    assert tool.strict is False
